# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"HyperParameters logic."


import abc

import six

from keras_tuner import protos
from keras_tuner import utils


@six.add_metaclass(abc.ABCMeta)
class Condition:
    """Abstract condition for a conditional hyperparameter.

    Subclasses of this object can be passed to a `HyperParameter` to specify
    that this condition must be met for the hyperparameter to be active for the
    `Trial`.
    """

    @abc.abstractmethod
    def is_active(self, values):
        """Whether this condition should be considered active.

        Determines whether this condition is true for the current `Trial`.

        Args:
            values: Dict. The active values for this `Trial`. Keys are the
                names of the hyperparameters.

        Returns:
            A boolean value of whether the condition is true.
        """
        raise NotImplementedError("Must be implemented in subclasses.")

    @abc.abstractmethod
    def __eq__(self, other):
        raise NotImplementedError("Must be implemented in subclasses.")

    @abc.abstractmethod
    def get_config(self):
        raise NotImplementedError("Must be implemented in subclasses.")

    @classmethod
    def from_config(cls, config):
        return cls(**config)  # pytype: disable=not-instantiable

    @classmethod
    def from_proto(cls, proto):
        kind = proto.WhichOneof("kind")
        if kind == "parent":
            parent = getattr(proto, kind)
            name = parent.name
            values = parent.values
            values = [getattr(v, v.WhichOneof("kind")) for v in values]
            return Parent(name=name, values=values)
        raise ValueError(f"Unrecognized condition of type: {kind}")


class Parent(Condition):
    """Condition checking a `HyperParameter`'s value is in a list of values.

    It specifies a condition that a `HyperParameter`'s value is in a list of
    values. It can be used as the condition to activate another
    `HyperParameter` for the `Trial`.

    Example:

    ```python
    a = Choice('model', ['linear', 'dnn'])
    condition = Parent(name='model', value=['dnn'])
    b = Int('num_layers', 5, 10, conditions=[condition])
    ```

    Args:
        name: A string, the name of the `HyperParameter` to use in the
            condition.
        values: A list of values of the `HyperParameter` to activate the
            condition.
    """

    def __init__(self, name, values):
        self.name = name

        # Standardize on str, int, float, bool.
        values = utils.to_list(values)
        first_val = values[0]
        if isinstance(first_val, six.string_types):
            values = [str(v) for v in values]
        elif isinstance(first_val, bool):
            # Bool check needs to be before integer check to prevent bool falls
            # into integer condition.
            pass
        elif isinstance(first_val, six.integer_types):
            values = [int(v) for v in values]
        elif not isinstance(first_val, float):
            raise TypeError(
                "Can contain only `int`, `float`, `str`, or "
                "`bool`, found values: " + str(values) + "with "
                "types: " + str(type(first_val))
            )
        self.values = values

    def is_active(self, values):
        return self.name in values and values[self.name] in self.values

    def __eq__(self, other):
        return (
            isinstance(other, Parent)
            and other.name == self.name
            and other.values == self.values
        )

    def get_config(self):
        return {"name": self.name, "values": self.values}

    def to_proto(self):
        print(self.values[0])
        if isinstance(self.values[0], six.string_types):
            values = [
                protos.get_proto().Value(string_value=v) for v in self.values
            ]
        elif isinstance(self.values[0], bool):
            values = [
                protos.get_proto().Value(boolean_value=v) for v in self.values
            ]
        elif isinstance(self.values[0], six.integer_types):
            values = [
                protos.get_proto().Value(int_value=v) for v in self.values
            ]
        else:
            values = [
                protos.get_proto().Value(float_value=v) for v in self.values
            ]

        return protos.get_proto().Condition(
            parent=protos.get_proto().Condition.Parent(
                name=self.name, values=values
            )
        )


OBJECTS = (
    Condition,
    Parent,
)
ALL_CLASSES = {cls.__name__: cls for cls in OBJECTS}


def deserialize(config):
    return utils.deserialize_keras_object(config, module_objects=ALL_CLASSES)


def serialize(obj):
    return utils.serialize_keras_object(obj)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for HyperXception Model."""

import numpy as np
import pytest
import tensorflow as tf

from keras_tuner.applications import xception
from keras_tuner.engine import hyperparameters as hp_module


@pytest.mark.parametrize("pooling", ["flatten", "avg", "max"])
def test_model_construction(pooling):
    hp = hp_module.HyperParameters()
    hp.Choice("pooling", [pooling])
    hypermodel = xception.HyperXception(input_shape=(128, 128, 3), classes=10)
    model = hypermodel.build(hp)
    assert hp.values["pooling"] == pooling
    assert model.layers
    assert model.name == "Xception"
    assert model.output_shape == (None, 10)
    model.train_on_batch(np.ones((1, 128, 128, 3)), np.ones((1, 10)))
    out = model.predict(np.ones((1, 128, 128, 3)))
    assert out.shape == (1, 10)


def test_hyperparameter_existence_and_defaults():
    hp = hp_module.HyperParameters()
    hypermodel = xception.HyperXception(input_shape=(256, 256, 3), classes=10)
    hypermodel.build(hp)
    assert hp.values == {
        "activation": "relu",
        "conv2d_num_filters": 64,
        "kernel_size": 3,
        "initial_strides": 2,
        "sep_num_filters": 256,
        "num_residual_blocks": 4,
        "pooling": "avg",
        "num_dense_layers": 1,
        "dropout_rate": 0.5,
        "dense_use_bn": True,
        "learning_rate": 1e-3,
    }


def test_include_top_false():
    hp = hp_module.HyperParameters()
    hypermodel = xception.HyperXception(
        input_shape=(256, 256, 3), classes=10, include_top=False
    )
    model = hypermodel.build(hp)
    assert not model.optimizer


def test_hyperparameter_override():
    hp = hp_module.HyperParameters()
    hp.Choice("pooling", ["flatten"])
    hp.Choice("num_dense_layers", [2])
    hypermodel = xception.HyperXception(input_shape=(256, 256, 3), classes=10)
    hypermodel.build(hp)
    assert hp.get("pooling") == "flatten"
    assert hp.get("num_dense_layers") == 2


def test_input_tensor():
    hp = hp_module.HyperParameters()
    inputs = tf.keras.Input((256, 256, 3))
    hypermodel = xception.HyperXception(input_tensor=inputs, include_top=False)
    model = hypermodel.build(hp)
    assert model.inputs == [inputs]


def test_activation_selu():
    hp = hp_module.HyperParameters()
    hp.values["activation"] = "selu"
    hypermodel = xception.HyperXception(input_shape=(256, 256, 3), classes=10)
    hypermodel.build(hp)


def test_no_classes_raise_error():
    with pytest.raises(ValueError, match="classes"):
        xception.HyperXception(input_shape=(256, 256, 3))


def test_no_input_shape_tensor_raise_error():
    with pytest.raises(ValueError, match="input_tensor"):
        xception.HyperXception(classes=10)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"HyperModel base class."

from keras_tuner import errors
from keras_tuner.api_export import keras_tuner_export


@keras_tuner_export("keras_tuner.HyperModel")
class HyperModel:
    """Defines a search space of models.

    A search space is a collection of models. The `build` function will build
    one of the models from the space using the given `HyperParameters` object.

    Users should subclass the `HyperModel` class to define their search spaces
    by overriding `build()`, which creates and returns the Keras model.
    Optionally, you may also override `fit()` to customize the training process
    of the model.

    Examples:

    In `build()`, you can create the model using the hyperparameters.

    ```python
    class MyHyperModel(kt.HyperModel):
        def build(self, hp):
            model = keras.Sequential()
            model.add(keras.layers.Dense(
                hp.Choice('units', [8, 16, 32]),
                activation='relu'))
            model.add(keras.layers.Dense(1, activation='relu'))
            model.compile(loss='mse')
            return model
    ```

    When overriding `HyperModel.fit()`, if you use `model.fit()` to train your
    model, which returns the training history, you can return it directly. You
    may use `hp` to specify any hyperparameters to tune.

    ```python
    class MyHyperModel(kt.HyperModel):
        def build(self, hp):
            ...

        def fit(self, hp, model, *args, **kwargs):
            return model.fit(
                *args,
                epochs=hp.Int("epochs", 5, 20),
                **kwargs)
    ```

    If you have a customized training process, you can return the objective
    value as a float.

    If you want to keep track of more metrics, you can return a dictionary of
    the metrics to track.

    ```python
    class MyHyperModel(kt.HyperModel):
        def build(self, hp):
            ...

        def fit(self, hp, model, *args, **kwargs):
            ...
            return {
                "loss": loss,
                "val_loss": val_loss,
                "val_accuracy": val_accuracy
            }
    ```

    Args:
        name: Optional string, the name of this HyperModel.
        tunable: Boolean, whether the hyperparameters defined in this
            hypermodel should be added to search space. If `False`, either the
            search space for these parameters must be defined in advance, or
            the default values will be used. Defaults to True.
    """

    def __init__(self, name=None, tunable=True):
        self.name = name
        self.tunable = tunable

        self._build = self.build
        self.build = self._build_wrapper

    def build(self, hp):
        """Builds a model.

        Args:
            hp: A `HyperParameters` instance.

        Returns:
            A model instance.
        """
        raise NotImplementedError

    def _build_wrapper(self, hp, *args, **kwargs):
        if not self.tunable:
            # Copy `HyperParameters` object so that new entries are not added
            # to the search space.
            hp = hp.copy()
        return self._build(hp, *args, **kwargs)

    def declare_hyperparameters(self, hp):
        pass

    def fit(self, hp, model, *args, **kwargs):
        """Train the model.

        Args:
            hp: HyperParameters.
            model: `keras.Model` built in the `build()` function.
            **kwargs: All arguments passed to `Tuner.search()` are in the
                `kwargs` here. It always contains a `callbacks` argument, which
                is a list of default Keras callback functions for model
                checkpointing, tensorboard configuration, and other tuning
                utilities. If `callbacks` is passed by the user from
                `Tuner.search()`, these default callbacks will be appended to
                the user provided list.

        Returns:
            A `History` object, which is the return value of `model.fit()`, a
            dictionary, or a float.

            If return a dictionary, it should be a dictionary of the metrics to
            track. The keys are the metric names, which contains the
            `objective` name. The values should be the metric values.

            If return a float, it should be the `objective` value.
        """
        return model.fit(*args, **kwargs)


class DefaultHyperModel(HyperModel):
    """Produces HyperModel from a model building function."""

    def __init__(self, build, name=None, tunable=True):
        super().__init__(name=name)
        self.build = build


def get_hypermodel(hypermodel):
    """Gets a HyperModel from a HyperModel or callable."""
    if hypermodel is None:
        return None

    if isinstance(hypermodel, HyperModel):
        return hypermodel

    if not callable(hypermodel):
        raise errors.FatalValueError(
            "The `hypermodel` argument should be either "
            "a callable with signature `build(hp)` returning a model, "
            "or an instance of `HyperModel`."
        )
    return DefaultHyperModel(hypermodel)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from tensorflow import keras

from keras_tuner import protos
from keras_tuner.engine import hyperparameters as hp_module


def test_hyperparameters():
    hp = hp_module.HyperParameters()
    assert hp.values == {}
    assert hp.space == []
    hp.Choice("choice", [1, 2, 3], default=2)
    assert hp.values == {"choice": 2}
    assert len(hp.space) == 1
    assert hp.space[0].name == "choice"
    hp.values["choice"] = 3
    assert hp.get("choice") == 3
    hp = hp.copy()
    assert hp.values == {"choice": 3}
    assert len(hp.space) == 1
    assert hp.space[0].name == "choice"
    with pytest.raises(KeyError, match="does not exist"):
        hp.get("wrong")


def test_name_collision():
    # TODO: figure out how name collision checks
    # should work.
    pass


def test_name_scope():
    hp = hp_module.HyperParameters()
    hp.Choice("choice", [1, 2, 3], default=2)
    with hp.name_scope("scope1"):
        hp.Choice("choice", [4, 5, 6], default=5)
        with hp.name_scope("scope2"):
            hp.Choice("choice", [7, 8, 9], default=8)
        hp.Int("range", min_value=0, max_value=10, step=1, default=0)
    assert hp.values == {
        "choice": 2,
        "scope1/choice": 5,
        "scope1/scope2/choice": 8,
        "scope1/range": 0,
    }


def test_parent_name():
    hp = hp_module.HyperParameters()
    hp.Choice("a", [1, 2, 3], default=2)
    b1 = hp.Int("b", 0, 10, parent_name="a", parent_values=1, default=5)
    b2 = hp.Int("b", 0, 100, parent_name="a", parent_values=2, default=4)
    assert b1 is None
    assert b2 == 4
    # Only active values appear in `values`.
    assert hp.values == {"a": 2, "b": 4}


def test_conditional_scope():
    hp = hp_module.HyperParameters()
    hp.Choice("choice", [1, 2, 3], default=2)
    # Assignment to a non-active conditional hyperparameter returns `None`.
    with hp.conditional_scope("choice", [1, 3]):
        child1 = hp.Choice("child_choice", [4, 5, 6])
    assert child1 is None
    # Retrieve a non-active hp, still none.
    with hp.conditional_scope("choice", [1, 3]):
        child1 = hp.Choice("child_choice", [4, 5, 6])
    assert child1 is None
    # Assignment to an active conditional hyperparameter returns the value.
    with hp.conditional_scope("choice", 2):
        child2 = hp.Choice("child_choice", [7, 8, 9])
    assert child2 == 7
    # Retrieve the value, still same value.
    with hp.conditional_scope("choice", 2):
        child2 = hp.Choice("child_choice", [7, 8, 9])
    assert child2 == 7
    # Only active values appear in `values`.
    assert hp.values == {"choice": 2, "child_choice": 7}

    with pytest.raises(ValueError, match="not defined"):
        with hp.conditional_scope("not_defined_hp", 2):
            hp.Choice("child_choice", [7, 8, 9])


def test_to_proto_unrecognized_hp_type():
    hps = hp_module.HyperParameters()
    hps._space.append(None)
    hps.Fixed("d", "3")

    with pytest.raises(ValueError, match="Unrecognized HP"):
        hp_module.HyperParameters.from_proto(hps.to_proto())


def test_to_proto_unrecognized_value_type():
    hps = hp_module.HyperParameters()
    hps.Fixed("d", "3")
    hps.values["d"] = None

    with pytest.raises(ValueError, match="Unrecognized value type"):
        hp_module.HyperParameters.from_proto(hps.to_proto())


def test_is_active_with_hp_name_and_hp():
    hp = hp_module.HyperParameters()
    hp.Choice("choice", [1, 2, 3], default=3)
    with hp.conditional_scope("choice", [1, 3]):
        hp.Choice("child_choice", [4, 5, 6])
    with hp.conditional_scope("choice", 2):
        hp.Choice("child_choice2", [7, 8, 9])

    # Custom oracle populates value for an inactive hp.
    hp.values["child_choice2"] = 7

    assert hp.is_active("child_choice")
    assert hp.is_active(hp._hps["child_choice"][0])

    assert not hp.is_active("child_choice2")
    assert not hp.is_active(hp._hps["child_choice2"][0])


def test_build_with_conditional_scope():
    def build_model(hp):
        model = hp.Choice("model", ["v1", "v2"])
        with hp.conditional_scope("model", "v1"):
            v1_params = {
                "layers": hp.Int("layers", 1, 3),
                "units": hp.Int("units", 16, 32),
            }
        with hp.conditional_scope("model", "v2"):
            v2_params = {
                "layers": hp.Int("layers", 2, 4),
                "units": hp.Int("units", 32, 64),
            }

        params = v1_params if model == "v1" else v2_params
        inputs = keras.Input(10)
        x = inputs
        for _ in range(params["layers"]):
            x = keras.layers.Dense(params["units"])(x)
        outputs = keras.layers.Dense(1)(x)
        model = keras.Model(inputs, outputs)
        model.compile("sgd", "mse")
        return model

    hp = hp_module.HyperParameters()
    build_model(hp)
    assert hp.values == {
        "model": "v1",
        "layers": 1,
        "units": 16,
    }


def test_error_when_hp_same_name_as_condition():
    hp = hp_module.HyperParameters()
    hp.Choice("a", [1, 2, 3], default=3)
    with pytest.raises(ValueError, match="cannot have the same name"):
        with hp.conditional_scope("a", [1, 3]):
            hp.Choice("a", [4, 5, 6], default=6)


def test_nested_conditional_scopes_and_name_scopes():
    hp = hp_module.HyperParameters()
    a = hp.Choice("a", [1, 2, 3], default=3)
    with hp.conditional_scope("a", [1, 3]):
        b = hp.Choice("b", [4, 5, 6], default=6)
        with hp.conditional_scope("b", 6):
            c = hp.Choice("c", [7, 8, 9])
            with hp.name_scope("d"):
                e = hp.Choice("e", [10, 11, 12])
    with hp.conditional_scope("a", 2):
        f = hp.Choice("f", [13, 14, 15])
        with hp.name_scope("g"):
            h = hp.Int("h", 0, 10)

    assert hp.values == {
        "a": 3,
        "b": 6,
        "c": 7,
        "d/e": 10,
    }
    # Assignment to an active conditional hyperparameter returns the value.
    assert a == 3
    assert b == 6
    assert c == 7
    assert e == 10
    # Assignment to a non-active conditional hyperparameter returns `None`.
    assert f is None
    assert h is None


def test_get_with_conditional_scopes():
    hp = hp_module.HyperParameters()
    hp.Choice("a", [1, 2, 3], default=2)
    assert hp.get("a") == 2
    with hp.conditional_scope("a", 2):
        hp.Fixed("b", 4)
        assert hp.get("b") == 4
        assert hp.get("a") == 2
    with hp.conditional_scope("a", 3):
        hp.Fixed("b", 5)
        assert hp.get("b") == 4

    # Value corresponding to the currently active condition is returned.
    assert hp.get("b") == 4


def test_merge_inactive_hp_with_conditional_scopes():
    hp = hp_module.HyperParameters()
    hp.Choice("a", [1, 2, 3], default=3)
    assert hp.get("a") == 3
    with hp.conditional_scope("a", 2):
        hp.Fixed("b", 4)

    hp2 = hp_module.HyperParameters()
    hp2.merge(hp)
    # only active hp should be included to values
    assert "a" in hp2.values
    assert "b" not in hp2.values


def test_merge():
    hp = hp_module.HyperParameters()
    hp.Int("a", 0, 100)
    hp.Fixed("b", 2)

    hp2 = hp_module.HyperParameters()
    hp2.Fixed("a", 3)
    hp.Int("c", 10, 100, default=30)

    hp.merge(hp2)

    assert hp.get("a") == 3
    assert hp.get("b") == 2
    assert hp.get("c") == 30

    hp3 = hp_module.HyperParameters()
    hp3.Fixed("a", 5)
    hp3.Choice("d", [1, 2, 3], default=1)

    hp.merge(hp3, overwrite=False)

    assert hp.get("a") == 3
    assert hp.get("b") == 2
    assert hp.get("c") == 30
    assert hp.get("d") == 1


def _sort_space(hps):
    space = hps.get_config()["space"]
    return sorted(space, key=lambda hp: hp["config"]["name"])


def test_hyperparameters_proto():
    hps = hp_module.HyperParameters()
    hps.Int("a", 1, 10, sampling="reverse_log", default=3)
    hps.Float("b", 2, 8, sampling="linear", default=4)
    hps.Choice("c", [1, 5, 10], ordered=False, default=5)
    hps.Fixed("d", "3")
    hps.Fixed("e", 3)
    hps.Fixed("f", 3.1)
    hps.Fixed("g", True)
    hps.Boolean("h")
    with hps.name_scope("d"):
        hps.Choice("e", [2.0, 4.5, 8.5], default=2.0)
        hps.Choice("f", ["1", "2"], default="1")
        with hps.conditional_scope("f", "1"):
            hps.Int("g", -10, 10, step=2, default=-2)

    new_hps = hp_module.HyperParameters.from_proto(hps.to_proto())
    assert _sort_space(hps) == _sort_space(new_hps)
    assert hps.values == new_hps.values


def test_hyperparameters_values_proto():
    values = protos.get_proto().HyperParameters.Values(
        values={
            "a": protos.get_proto().Value(int_value=1),
            "b": protos.get_proto().Value(float_value=2.0),
            "c": protos.get_proto().Value(string_value="3"),
        }
    )

    # When only values are provided, each param is created as `Fixed`.
    hps = hp_module.HyperParameters.from_proto(values)
    assert hps.values == {"a": 1, "b": 2.0, "c": "3"}


def test_dict_methods():
    hps = hp_module.HyperParameters()
    hps.Int("a", 0, 10, default=3)
    hps.Choice("b", [1, 2], default=2)
    with hps.conditional_scope("b", 1):
        hps.Float("c", -10, 10, default=3)
        # Don't allow access of a non-active param within its scope.
        with pytest.raises(ValueError, match="is currently inactive"):
            hps["c"]
    with hps.conditional_scope("b", 2):
        hps.Float("c", -30, -20, default=-25)

    assert hps["a"] == 3
    assert hps["b"] == 2
    # Ok to access 'c' here since there is an active 'c'.
    assert hps["c"] == -25
    with pytest.raises(KeyError, match="does not exist"):
        hps["d"]

    assert "a" in hps
    assert "b" in hps
    assert "c" in hps
    assert "d" not in hps


def test_return_populated_value_for_new_hp():
    hp = hp_module.HyperParameters()

    hp.values["hp_name"] = "hp_value"
    assert (
        hp.Choice(
            "hp_name",
            ["hp_value", "hp_value_default"],
            default="hp_value_default",
        )
        == "hp_value"
    )


def test_return_default_value_if_not_populated():
    hp = hp_module.HyperParameters()

    assert (
        hp.Choice(
            "hp_name",
            ["hp_value", "hp_value_default"],
            default="hp_value_default",
        )
        == "hp_value_default"
    )


def test_serialize_deserialize_hyperparameters():
    hp = hp_module.HyperParameters()
    hp.Int("temp", 1, 5)
    hp = hp_module.deserialize(hp_module.serialize(hp))
    assert len(hp.space) == 1


def test_int_log_without_step_random_sample():
    hp = hp_module.HyperParameters()
    hp.Int("rg", min_value=2, max_value=32, sampling="log")
    hp.space[0].random_sample()



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Basic exhaustive search tuner."


import collections
import copy

from keras_tuner.api_export import keras_tuner_export
from keras_tuner.engine import oracle as oracle_module
from keras_tuner.engine import trial as trial_module
from keras_tuner.engine import tuner as tuner_module


class LinkedList:
    """A simplified linked list with limited supported operations.

    It doesn't copy any data pass to it but directly refer to it.
    """

    def __init__(self):
        # _memory is a list to store data.
        # Its index is the address for the linked list.
        # index to data
        self._memory = []
        self._data_to_index = {}
        # index to index
        self._next_index = collections.defaultdict(lambda: None)
        self._last_index = None

    def insert(self, data, data_pre=None):
        """Insert data after another data.

        `data` is inserted after `data_pre` in the linked list.

        Args:
            data: The data to insert.
            data_pre: Optional. The data marking the insertion location. If left
                unspecified, the data will be appended to the rear of the linked
                list.
        """
        self._memory.append(data)
        new_index = len(self._memory) - 1
        self._data_to_index[data] = new_index

        index = (
            self._last_index
            if data_pre is None
            else self._data_to_index[data_pre]
        )

        self._next_index[new_index] = self._next_index[index]
        self._next_index[index] = new_index

        # Update self._last_index.
        while self._next_index[self._last_index] is not None:
            self._last_index = self._next_index[self._last_index]

    def next(self, data):
        """Get the next data for a given data.

        Args:
            data: The data used to get its next data in the linked list.

        Returns:
            The next data if exists. Otherwise, return None.
        """
        index = self._data_to_index[data]
        next_index = self._next_index[index]
        if next_index is None:
            return None
        return self._memory[next_index]


@keras_tuner_export("keras_tuner.oracles.GridSearchOracle")
class GridSearchOracle(oracle_module.Oracle):
    """Grid search oracle.

    Args:
        objective: A string, `keras_tuner.Objective` instance, or a list of
            `keras_tuner.Objective`s and strings. If a string, the direction of
            the optimization (min or max) will be inferred. If a list of
            `keras_tuner.Objective`, we will minimize the sum of all the
            objectives to minimize subtracting the sum of all the objectives to
            maximize. The `objective` argument is optional when
            `Tuner.run_trial()` or `HyperModel.fit()` returns a single float as
            the objective to minimize.
        max_trials: Optional integer, the total number of trials (model
            configurations) to test at most. Note that the oracle may interrupt
            the search before `max_trial` models have been tested if the search
            space has been exhausted. If left unspecified, it runs till the
            search space is exhausted.
        seed: Optional integer, the random seed.
        hyperparameters: Optional `HyperParameters` instance. Can be used to
            override (or register in advance) hyperparameters in the search
            space.
        tune_new_entries: Boolean, whether hyperparameter entries that are
            requested by the hypermodel but that were not specified in
            `hyperparameters` should be added to the search space, or not. If
            not, then the default value for these parameters will be used.
            Defaults to True.
        allow_new_entries: Boolean, whether the hypermodel is allowed to
            request hyperparameter entries not listed in `hyperparameters`.
            Defaults to True.
        max_retries_per_trial: Integer. Defaults to 0. The maximum number of
            times to retry a `Trial` if the trial crashed or the results are
            invalid.
        max_consecutive_failed_trials: Integer. Defaults to 3. The maximum
            number of consecutive failed `Trial`s. When this number is reached,
            the search will be stopped. A `Trial` is marked as failed when none
            of the retries succeeded.
    """

    def __init__(
        self,
        objective=None,
        max_trials=None,
        seed=None,
        hyperparameters=None,
        allow_new_entries=True,
        tune_new_entries=True,
        max_retries_per_trial=0,
        max_consecutive_failed_trials=3,
    ):
        super().__init__(
            objective=objective,
            max_trials=max_trials,
            hyperparameters=hyperparameters,
            tune_new_entries=tune_new_entries,
            allow_new_entries=allow_new_entries,
            seed=seed,
            max_retries_per_trial=max_retries_per_trial,
            max_consecutive_failed_trials=max_consecutive_failed_trials,
        )
        # List of trial_id sorting in ascending alphabetical order of their hp
        # values.
        self._ordered_ids = LinkedList()
        # Queue of trial_ids pending to find their next combinations.
        self._populate_next = []

    def populate_space(self, trial_id):
        """Fill the hyperparameter space with values.

        Args:
            trial_id: A string, the ID for this Trial.

        Returns:
            A dictionary with keys "values" and "status", where "values" is
            a mapping of parameter names to suggested values, and "status"
            should be one of "RUNNING" (the trial can start normally), "IDLE"
            (the oracle is waiting on something and cannot create a trial), or
            "STOPPED" (the oracle has finished searching and no new trial should
            be created).
        """
        values = None

        # See if this is the first trial.
        if len(self.start_order) == 0:
            # Use all default values for the first trial.
            self._ordered_ids.insert(trial_id)
            hps = self.get_space()
            values = {
                hp.name: hp.default
                for hp in self.get_space().space
                if hps.is_active(hp)
            }
            # Although the trial is not finished, we still push it into
            # _populate_next to quickly generate values for the first few trials
            # for multiple workers. The same trial_id will be pushed into
            # _populate_next again when the trial is finished just in case of
            # new hps appeared during the trial.
            self._populate_next.append(trial_id)

        # Pick tried values to create its next combination if not tried.
        while len(self._populate_next) > 0 and values is None:
            old_trial_id = self._populate_next.pop(0)

            # Create its immediate next combination.
            old_values = self.trials[old_trial_id].hyperparameters.values
            new_values = self._get_next_combination(old_values)
            if new_values is None:
                continue

            # Skip if tried next combination.
            next_id = self._ordered_ids.next(old_trial_id)
            if next_id is not None:
                next_values = self.trials[next_id].hyperparameters.values
                if self._compare(new_values, next_values) >= 0:
                    continue

            self._ordered_ids.insert(trial_id, old_trial_id)

            values = new_values

        if values is not None:
            return {
                "status": trial_module.TrialStatus.RUNNING,
                "values": values,
            }

        # Wait for the ongoing trials to finish when the values queue is empty
        # in case of any new hp discovered.
        if len(self.ongoing_trials) > 0:
            return {"status": trial_module.TrialStatus.IDLE, "values": None}

        # Reaching this point means ongoing_trial, values, populate_next
        # are all empty.
        return {"status": trial_module.TrialStatus.STOPPED, "values": None}

    def _compare(self, a, b):
        """Compare two `HyperParameters`' values.

        The smallest index where a differs from b decides which one is larger.
        In the values of one `HyperParameter`, the default value is the
        smallest. The rest are sorted according to their order in
        `HyperParameter.values`.  If one value is the prefix of another, the
        longer one is larger.

        Args:
            a: Dict. HyperParameters values. Only active values are included.
            b: Dict. HyperParameters values. Only active values are included.

        Returns:
            -1 if a < b, 0 if a == b, 1 if a > b.
        """
        hps = self.get_space()
        for hp in hps.space:
            # The hp is not active in neither a or b.
            # Whether it is active should be the same in a and b,
            # or the loop have stopped at the parent values which are different.
            if hp.name not in a:
                continue

            if a[hp.name] == b[hp.name]:
                continue

            # Get a ordered list of the values of the hp.
            value_list = list(hp.values)
            if hp.default in value_list:
                value_list.remove(hp.default)
            value_list.insert(0, hp.default)

            index_a = value_list.index(a[hp.name])
            index_b = value_list.index(b[hp.name])
            return -1 if index_a < index_b else 1

        return 0

    def _get_next_combination(self, values):
        """Get the next value combination to try.

        Given the last trial's values dictionary, this method retrieves the next
        hyperparameter values to try. As it requires the last trial's
        values as input, it should not be called on the first trial. The first
        trial will always use default hp values.

        This oracle iterates over the search space entirely deterministically.

        When a new hp appears in a trial, the first value tried for that hp
        will be its default value.

        Args:
            values: Dict. The keys are hp names. The values are the hp values
                from the last trial.

        Returns:
            Dict or None. The next possible value combination for the
            hyperparameters. If no next combination exist (values is the last
            combination), it returns None. The return values only include the
            active ones.
        """

        hps = self.get_space()
        all_values = {}
        for hp in hps.space:
            value_list = list(hp.values)
            if hp.default in value_list:
                value_list.remove(hp.default)
            # Put the default value first.
            all_values[hp.name] = [hp.default] + value_list
        default_values = {hp.name: hp.default for hp in hps.space}
        hps.values = copy.deepcopy(values)

        bumped_value = False

        # Iterate in reverse order so that we can change the value under
        # conditional scope first instead of change the condition value first.
        for hp in reversed(hps.space):
            name = hp.name
            # Bump up the hp value if possible and active.
            if hps.is_active(hp):
                value = hps.values[name]
                if value != all_values[name][-1]:
                    index = all_values[name].index(value) + 1
                    hps.values[name] = all_values[name][index]
                    bumped_value = True
                    break
            # Otherwise, reset to its first value.
            hps.values[name] = default_values[name]

        hps.ensure_active_values()
        return hps.values if bumped_value else None

    @oracle_module.synchronized
    def end_trial(self, trial):
        super().end_trial(trial)
        # It is OK for a trial_id to be pushed into _populate_next multiple
        # times. It will be skipped during _populate_space if its next
        # combination has been tried.

        # For not blocking _populate_space, we push it regardless of the status.
        self._populate_next.append(trial.trial_id)


@keras_tuner_export(["keras_tuner.GridSearch", "keras_tuner.tuners.GridSearch"])
class GridSearch(tuner_module.Tuner):
    """The grid search tuner.

    This tuner iterates over all possible
    hyperparameter combinations.

    For example, with:

    ```py
    optimizer = hp.Choice("model_name", values=["sgd", "adam"])
    learning_rate = hp.Choice("learning_rate", values=[0.01, 0.1])
    ```

    This tuner will cover the following combinations:
    `["sgd", 0.01], ["sgd", 0.1], ["adam", 0.01] ["adam", 0.1]`.

    For the following hyperparameter types, GridSearch will not exhaust all
    possible values:

    * `hp.Float()` when `step` is left unspecified.
    * `hp.Int()` with `sampling` set to `"log"` or `"reverse_log"`, and `step`
        is left unspecified.

    For these cases, KerasTuner will pick 10 samples in the range evenly by
    default. To configure the granularity of sampling for `hp.Float()` and
    `hp.Int()`, please use the `step` argument in their initializers.

    Args:
        hypermodel: Instance of `HyperModel` class (or callable that takes
            hyperparameters and returns a Model instance). It is optional when
            `Tuner.run_trial()` is overridden and does not use
            `self.hypermodel`.
        objective: A string, `keras_tuner.Objective` instance, or a list of
            `keras_tuner.Objective`s and strings. If a string, the direction of
            the optimization (min or max) will be inferred. If a list of
            `keras_tuner.Objective`, we will minimize the sum of all the
            objectives to minimize subtracting the sum of all the objectives to
            maximize. The `objective` argument is optional when
            `Tuner.run_trial()` or `HyperModel.fit()` returns a single float as
            the objective to minimize.
        max_trials: Optional integer, the total number of trials (model
            configurations) to test at most. Note that the oracle may interrupt
            the search before `max_trial` models have been tested if the search
            space has been exhausted. If left unspecified, it runs till the
            search space is exhausted.
        seed: Optional integer, the random seed.
        hyperparameters: Optional `HyperParameters` instance. Can be used to
            override (or register in advance) hyperparameters in the search
            space.
        tune_new_entries: Boolean, whether hyperparameter entries that are
            requested by the hypermodel but that were not specified in
            `hyperparameters` should be added to the search space, or not. If
            not, then the default value for these parameters will be used.
            Defaults to True.
        allow_new_entries: Boolean, whether the hypermodel is allowed to
            request hyperparameter entries not listed in `hyperparameters`.
            Defaults to True.
        max_retries_per_trial: Integer. Defaults to 0. The maximum number of
            times to retry a `Trial` if the trial crashed or the results are
            invalid.
        max_consecutive_failed_trials: Integer. Defaults to 3. The maximum
            number of consecutive failed `Trial`s. When this number is reached,
            the search will be stopped. A `Trial` is marked as failed when none
            of the retries succeeded.
        **kwargs: Keyword arguments relevant to all `Tuner` subclasses.
            Please see the docstring for `Tuner`.
    """

    def __init__(
        self,
        hypermodel=None,
        objective=None,
        max_trials=None,
        seed=None,
        hyperparameters=None,
        tune_new_entries=True,
        allow_new_entries=True,
        max_retries_per_trial=0,
        max_consecutive_failed_trials=3,
        **kwargs,
    ):
        self.seed = seed
        oracle = GridSearchOracle(
            objective=objective,
            max_trials=max_trials,
            seed=seed,
            hyperparameters=hyperparameters,
            tune_new_entries=tune_new_entries,
            allow_new_entries=allow_new_entries,
            max_retries_per_trial=max_retries_per_trial,
            max_consecutive_failed_trials=max_consecutive_failed_trials,
        )
        super().__init__(oracle, hypermodel, **kwargs)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np
import pytest
import tensorflow as tf

import keras_tuner
from keras_tuner.engine import hyperparameters as hp_module
from keras_tuner.engine import trial as trial_module
from keras_tuner.tuners import bayesian as bo_module


def build_model(hp):
    model = tf.keras.Sequential()
    model.add(tf.keras.layers.Flatten(input_shape=(2, 2)))
    for i in range(3):
        model.add(
            tf.keras.layers.Dense(
                units=hp.Int(f"units_{str(i)}", 2, 4, 2), activation="relu"
            )
        )

    model.add(tf.keras.layers.Dense(2, activation="softmax"))
    model.compile(
        optimizer=tf.keras.optimizers.Adam(
            hp.Choice("learning_rate", [1e-2, 1e-3, 1e-4])
        ),
        loss="sparse_categorical_crossentropy",
        metrics=["accuracy"],
    )
    return model


def test_scipy_not_install_error(tmp_path):
    scipy_module = keras_tuner.tuners.bayesian.scipy
    keras_tuner.tuners.bayesian.scipy = None

    with pytest.raises(ImportError, match="Please install scipy"):
        keras_tuner.BayesianOptimization(
            hypermodel=build_model,
            directory=tmp_path,
        )

    keras_tuner.tuners.bayesian.scipy = scipy_module


def test_sklearn_not_install_error(tmp_path):
    sklearn_module = keras_tuner.tuners.bayesian.sklearn
    keras_tuner.tuners.bayesian.sklearn = None

    with pytest.raises(ImportError, match="Please install scikit-learn"):
        keras_tuner.BayesianOptimization(
            hypermodel=build_model,
            directory=tmp_path,
        )

    keras_tuner.tuners.bayesian.sklearn = sklearn_module


def test_bayesian_oracle(tmp_path):
    hps = hp_module.HyperParameters()
    hps.Choice("a", [1, 2], default=1)
    hps.Int("b", 3, 10, default=3)
    hps.Float("c", 0, 1, 0.1, default=0)
    hps.Fixed("d", 7)
    hps.Choice("e", [9, 0], default=9)
    oracle = bo_module.BayesianOptimizationOracle(
        objective=keras_tuner.Objective("score", "max"),
        max_trials=20,
        num_initial_points=2,
        hyperparameters=hps,
    )
    oracle._set_project_dir(tmp_path, "untitled")
    for i in range(5):
        trial = oracle.create_trial(str(i))
        oracle.update_trial(trial.trial_id, {"score": i})
        trial.status = "COMPLETED"
        oracle.end_trial(trial)


def test_bayesian_oracle_with_zero_y(tmp_path):
    hps = hp_module.HyperParameters()
    hps.Choice("a", [1, 2], default=1)
    hps.Int("b", 3, 10, default=3)
    hps.Float("c", 0, 1, 0.1, default=0)
    hps.Fixed("d", 7)
    hps.Choice("e", [9, 0], default=9)
    oracle = bo_module.BayesianOptimizationOracle(
        objective=keras_tuner.Objective("score", "max"),
        max_trials=20,
        num_initial_points=2,
        hyperparameters=hps,
    )
    oracle._set_project_dir(tmp_path, "untitled")
    for i in range(5):
        trial = oracle.create_trial(str(i))
        oracle.update_trial(trial.trial_id, {"score": 0})
        trial.status = "COMPLETED"
        oracle.end_trial(trial)


def test_bayesian_dynamic_space(tmp_path):
    hps = hp_module.HyperParameters()
    hps.Choice("a", [1, 2], default=1)
    oracle = bo_module.BayesianOptimizationOracle(
        objective="val_acc", max_trials=20, num_initial_points=10
    )
    oracle._set_project_dir(tmp_path, "untitled")
    oracle.hyperparameters = hps
    for i in range(10):
        oracle.populate_space(str(i))
    hps.Int("b", 3, 10, default=3)
    assert "b" in oracle.populate_space("1_0")["values"]
    hps.Float("c", 0, 1, 0.1, default=0)
    assert "c" in oracle.populate_space("1_1")["values"]
    hps.Fixed("d", 7)
    assert "d" in oracle.populate_space("1_2")["values"]
    hps.Choice("e", [9, 0], default=9)
    assert "e" in oracle.populate_space("1_3")["values"]


def test_bayesian_save_reload(tmp_path):
    hps = hp_module.HyperParameters()
    hps.Choice("a", [1, 2], default=1)
    hps.Choice("b", [3, 4], default=3)
    hps.Choice("c", [5, 6], default=5)
    hps.Choice("d", [7, 8], default=7)
    hps.Choice("e", [9, 0], default=9)
    oracle = bo_module.BayesianOptimizationOracle(
        objective=keras_tuner.Objective("score", "max"),
        max_trials=20,
        hyperparameters=hps,
    )
    oracle._set_project_dir(tmp_path, "untitled")

    for _ in range(3):
        trial = oracle.create_trial("tuner_id")
        oracle.update_trial(trial.trial_id, {"score": 1.0})
        trial.status = "COMPLETED"
        oracle.end_trial(trial)

    oracle.save()
    oracle = bo_module.BayesianOptimizationOracle(
        objective=keras_tuner.Objective("score", "max"),
        max_trials=20,
        hyperparameters=hps,
    )
    oracle._set_project_dir(tmp_path, "untitled")
    oracle.reload()

    for _ in range(3):
        trial = oracle.create_trial("tuner_id")
        oracle.update_trial(trial.trial_id, {"score": 1.0})
        trial.status = "COMPLETED"
        oracle.end_trial(trial)

    assert len(oracle.trials) == 6


def test_bayesian_optimization_tuner(tmp_path):
    tuner = bo_module.BayesianOptimization(
        build_model, objective="val_accuracy", max_trials=15, directory=tmp_path
    )
    assert isinstance(tuner.oracle, bo_module.BayesianOptimizationOracle)


def test_bayesian_optimization_tuner_set_alpha_beta(tmp_path):
    tuner = bo_module.BayesianOptimization(
        build_model,
        alpha=1e-4,
        beta=2.6,
        objective="val_accuracy",
        max_trials=15,
        directory=tmp_path,
    )
    assert isinstance(tuner.oracle, bo_module.BayesianOptimizationOracle)


def test_save_before_result(tmp_path):
    hps = hp_module.HyperParameters()
    hps.Choice("a", [1, 2], default=1)
    hps.Int("b", 3, 10, default=3)
    hps.Float("c", 0, 1, 0.1, default=0)
    hps.Fixed("d", 7)
    hps.Choice("e", [9, 0], default=9)
    oracle = bo_module.BayesianOptimizationOracle(
        objective=keras_tuner.Objective("score", "max"),
        max_trials=10,
        hyperparameters=hps,
    )
    oracle._set_project_dir(tmp_path, "untitled")
    oracle.populate_space(str(1))
    oracle.save()


def test_bayesian_oracle_maximize(tmp_path):
    hps = hp_module.HyperParameters()
    hps.Int("a", -100, 100)

    oracle = bo_module.BayesianOptimizationOracle(
        objective=keras_tuner.Objective("score", direction="max"),
        max_trials=20,
        hyperparameters=hps,
        num_initial_points=2,
    )
    oracle._set_project_dir(tmp_path, "untitled")

    # Make examples with high 'a' and high score.
    for i in range(5):
        trial = trial_module.Trial(hyperparameters=hps.copy())
        trial.hyperparameters.values["a"] = 10 * i
        trial.score = i
        trial.status = "COMPLETED"
        oracle.trials[trial.trial_id] = trial

    # Make examples with low 'a' and low score
    for i in range(5):
        trial = trial_module.Trial(hyperparameters=hps.copy())
        trial.hyperparameters.values["a"] = -10 * i
        trial.score = -i
        trial.status = "COMPLETED"
        oracle.trials[trial.trial_id] = trial

    trial = oracle.create_trial("tuner0")
    assert trial.status == "RUNNING"
    # Assert that the oracle suggests hps it thinks will maximize.
    assert trial.hyperparameters.get("a") > 0


def test_hyperparameters_added(tmp_path):
    hps = hp_module.HyperParameters()
    hps.Int("a", -100, 100)

    oracle = bo_module.BayesianOptimizationOracle(
        objective=keras_tuner.Objective("score", direction="max"),
        max_trials=20,
        hyperparameters=hps,
        num_initial_points=2,
    )
    oracle._set_project_dir(tmp_path, "untitled")

    # Populate initial trials.
    for i in range(10):
        trial = trial_module.Trial(hyperparameters=hps.copy())
        trial.hyperparameters.values["a"] = 10 * i
        trial.score = i
        trial.status = "COMPLETED"
        oracle.trials[trial.trial_id] = trial

    # A new trial discovered a new hp and synced to oracle.hyperparameters.
    new_hps = hp_module.HyperParameters()
    new_hps.Float("b", 3.2, 6.4, step=0.2, default=3.6)
    new_hps.Boolean("c", default=True)
    oracle.update_space(new_hps)

    # Make a new trial, it should have b set.
    trial = oracle.create_trial("tuner0")
    assert trial.status == "RUNNING"
    assert "b" in trial.hyperparameters.values
    assert "c" in trial.hyperparameters.values


def test_step_respected(tmp_path):
    hps = hp_module.HyperParameters()
    hps.Float("c", 0, 10, step=3)
    oracle = bo_module.BayesianOptimizationOracle(
        objective=keras_tuner.Objective("score", direction="max"),
        max_trials=20,
        hyperparameters=hps,
        num_initial_points=2,
    )
    oracle._set_project_dir(tmp_path, "untitled")

    # Populate initial trials.
    for i in range(10):
        trial = trial_module.Trial(hyperparameters=hps.copy())
        trial.hyperparameters.values["c"] = 3.0
        trial.score = i
        trial.status = "COMPLETED"
        oracle.trials[trial.trial_id] = trial

    trial = oracle.create_trial("tuner0")
    # Check that oracle respects the `step` param.
    assert trial.hyperparameters.get("c") in {0, 3, 6, 9}


def test_float_optimization(tmp_path):
    class PolynomialTuner(keras_tuner.engine.base_tuner.BaseTuner):
        def run_trial(self, trial):
            hp = trial.hyperparameters
            return -1 * hp["a"] ** 3 + hp["b"] ** 3 + hp["c"] - abs(hp["d"])

    hps = hp_module.HyperParameters()
    hps.Float("a", -1, 1)
    hps.Float("b", -1, 1)
    hps.Float("c", -1, 1)
    hps.Float("d", -1, 1)

    tuner = PolynomialTuner(
        oracle=keras_tuner.oracles.BayesianOptimizationOracle(
            objective=keras_tuner.Objective("score", "max"),
            hyperparameters=hps,
            max_trials=50,
        ),
        directory=tmp_path,
    )

    tuner.search()

    atol, rtol = 1e-1, 1e-1
    best_trial = tuner.oracle.get_best_trials()[0]
    best_hps = best_trial.hyperparameters

    assert np.isclose(best_trial.score, 3, atol=atol, rtol=rtol)
    assert np.isclose(best_hps["a"], -1, atol=atol, rtol=rtol)
    assert np.isclose(best_hps["b"], 1, atol=atol, rtol=rtol)
    assert np.isclose(best_hps["c"], 1, atol=atol, rtol=rtol)
    assert np.isclose(best_hps["d"], 0, atol=atol, rtol=rtol)


def test_distributed_optimization(tmp_path):
    hps = hp_module.HyperParameters()
    hps.Int("a", 0, 10)
    hps.Float("b", -1, 1, step=0.1)
    hps.Float("c", 1e-5, 1e-2, sampling="log")

    def evaluate(hp):
        # Minimum at a=4, b=1, c=1e-3 with score=-1
        return abs(hp["a"] - 4) - hp["b"] + 0.1 * abs(3 + math.log(hp["c"], 10))

    oracle = bo_module.BayesianOptimizationOracle(
        objective=keras_tuner.Objective("score", "min"),
        hyperparameters=hps,
        max_trials=60,
    )
    oracle._set_project_dir(tmp_path, "untitled")

    tuners = 4

    for _ in range(10):
        trials = []
        for i in range(tuners):
            trial = oracle.create_trial(f"tuner_{str(i)}")
            trials.append(trial)
        for trial in trials:
            oracle.update_trial(
                trial.trial_id, {"score": evaluate(trial.hyperparameters)}
            )
        for trial in trials:
            trial.status = "COMPLETED"
            oracle.end_trial(trial)

    atol, rtol = 1e-1, 1e-1
    best_trial = oracle.get_best_trials()[0]
    best_hps = best_trial.hyperparameters

    # The minimum is not always found but it is always close.
    assert best_trial.score < -0.8, best_hps.values
    assert np.isclose(best_hps["a"], 4, atol=atol, rtol=rtol)
    assert np.isclose(best_hps["b"], 1, atol=atol, rtol=rtol)

    # For log-scale param, just check that the order of magnitude is correct.
    log_best_c = math.log(best_hps["c"], 10)
    assert log_best_c > -4 and log_best_c < -2


def test_interleaved_distributed_optimization(tmp_path):
    hps = hp_module.HyperParameters()
    hps.Float("a", -1, 1)
    hps.Float("b", -1, 1)
    hps.Float("c", -1, 1)
    hps.Float("d", -1, 1)

    def evaluate(hp):
        # Minimum at a=4, b=1, c=1e-3 with score=-1
        return -1 * hp["a"] ** 3 + hp["b"] ** 3 + hp["c"] - abs(hp["d"])

    oracle = bo_module.BayesianOptimizationOracle(
        objective=keras_tuner.Objective("score", "min"),
        hyperparameters=hps,
        max_trials=60,
        num_initial_points=2,
    )
    oracle._set_project_dir(tmp_path, "untitled")

    # Run 4 trials on 2 tuners

    # Start both tuners at the same time
    trial_1 = oracle.create_trial("tuner_0")
    trial_2 = oracle.create_trial("tuner_1")

    # tuner_0 finishes trial_1 before tuner_1 finishes
    oracle.update_trial(
        trial_1.trial_id, {"score": evaluate(trial_1.hyperparameters)}
    )
    trial_1.status = "COMPLETED"
    oracle.end_trial(trial_1)

    # tuner_0 request a new trial (trial_3)
    trial_3 = oracle.create_trial("tuner_0")

    # tuner_1 finishes trial_2
    oracle.update_trial(
        trial_2.trial_id, {"score": evaluate(trial_2.hyperparameters)}
    )
    trial_2.status = "COMPLETED"
    oracle.end_trial(trial_2)

    # tuner_1 requests the final new trial (trial_4)
    # the Bayesian optimizer will use ongoing trial_3 to hallucinate
    trial_4 = oracle.create_trial("tuner_1")

    # tuner_0 finishes trial_3
    oracle.update_trial(
        trial_3.trial_id, {"score": evaluate(trial_3.hyperparameters)}
    )
    trial_3.status = "COMPLETED"
    oracle.end_trial(trial_3)

    # tuner_1 finishes trial_4
    oracle.update_trial(
        trial_4.trial_id, {"score": evaluate(trial_4.hyperparameters)}
    )
    trial_4.status = "COMPLETED"
    oracle.end_trial(trial_4)

    assert True


def test_exhausted_search_space(tmp_path):
    class MyTuner(bo_module.BayesianOptimization):
        def run_trial(self, trial, *args, **kwargs):
            hp = trial.hyperparameters
            hp.Boolean("boolean")
            hp.Boolean("boolean2")
            return [np.random.rand()]

    tuner = MyTuner(
        alpha=1e-4,
        beta=2.6,
        max_trials=15,
        directory=tmp_path,
    )
    tuner.search()
    assert len(tuner.oracle.trials) == 4


def test_skip_failed_trials(tmp_path):
    class MyTuner(bo_module.BayesianOptimization):
        def run_trial(self, trial, *args, **kwargs):
            hp = trial.hyperparameters
            hp.Boolean("boolean")
            hp.Boolean("boolean2")
            hp.Boolean("boolean3")
            hp.Boolean("boolean4")
            if len(self.oracle.start_order) == 1:
                raise keras_tuner.errors.FailedTrialError()
            return [np.random.rand()]

    tuner = MyTuner(
        alpha=1e-4,
        beta=2.6,
        max_trials=15,
        directory=tmp_path,
    )
    tuner.search()
    assert any(
        map(lambda x: x.status == "FAILED", tuner.oracle.trials.values())
    )



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import os

import numpy as np
import pytest
from tensorflow import keras

from keras_tuner.engine import objective as obj_module
from keras_tuner.engine import tuner_utils
from keras_tuner.tuners import gridsearch


def test_save_best_epoch_with_single_objective(tmp_path):
    objective = obj_module.create_objective("val_loss")
    filepath = os.path.join(tmp_path, "saved_weights")
    callback = tuner_utils.SaveBestEpoch(objective, filepath)

    model = keras.Sequential([keras.layers.Dense(1)])
    model.compile(loss="mse")
    val_x = np.random.rand(10, 10)
    val_y = np.random.rand(10, 10)
    history = model.fit(
        x=np.random.rand(10, 10),
        y=np.random.rand(10, 1),
        validation_data=(val_x, val_y),
        epochs=10,
        callbacks=[callback],
    )

    model.load_weights(filepath)

    assert min(history.history["val_loss"]) == model.evaluate(val_x, val_y)


def test_save_best_epoch_with_multi_objective(tmp_path):
    objective = obj_module.create_objective(["val_loss", "val_mae"])
    filepath = os.path.join(tmp_path, "saved_weights")
    callback = tuner_utils.SaveBestEpoch(objective, filepath)

    model = keras.Sequential([keras.layers.Dense(1)])
    model.compile(loss="mse", metrics=["mae"])
    val_x = np.random.rand(10, 10)
    val_y = np.random.rand(10, 10)
    history = model.fit(
        x=np.random.rand(10, 10),
        y=np.random.rand(10, 1),
        validation_data=(val_x, val_y),
        epochs=10,
        callbacks=[callback],
    )

    model.load_weights(filepath)

    assert min(history.history["val_loss"]) + min(
        history.history["val_mae"]
    ) == sum(model.evaluate(val_x, val_y))


def test_save_best_epoch_with_default_objective(tmp_path):
    objective = obj_module.create_objective(None)
    filepath = os.path.join(tmp_path, "saved_weights")
    callback = tuner_utils.SaveBestEpoch(objective, filepath)

    model = keras.Sequential([keras.layers.Dense(1)])
    model.compile(loss="mse")
    val_x = np.random.rand(10, 10)
    val_y = np.random.rand(10, 10)
    history = model.fit(
        x=np.random.rand(10, 10),
        y=np.random.rand(10, 1),
        validation_data=(val_x, val_y),
        epochs=10,
        callbacks=[callback],
    )

    model.load_weights(filepath)

    assert history.history["val_loss"][-1] == model.evaluate(val_x, val_y)


def test_convert_to_metrics_with_history():
    model = keras.Sequential([keras.layers.Dense(1)])
    model.compile(loss="mse", metrics=["mae"])
    val_x = np.random.rand(10, 10)
    val_y = np.random.rand(10, 10)
    history = model.fit(
        x=np.random.rand(10, 10),
        y=np.random.rand(10, 1),
        validation_data=(val_x, val_y),
    )

    results = tuner_utils.convert_to_metrics_dict(
        history,
        obj_module.Objective("val_loss", "min"),
    )
    assert all(key in results for key in ["loss", "val_loss", "mae", "val_mae"])


def test_convert_to_metrics_with_float():
    assert tuner_utils.convert_to_metrics_dict(
        0.1,
        obj_module.Objective("val_loss", "min"),
    ) == {"val_loss": 0.1}


def test_convert_to_metrics_with_dict():
    assert tuner_utils.convert_to_metrics_dict(
        {"loss": 0.2, "val_loss": 0.1},
        obj_module.Objective("val_loss", "min"),
    ) == {"loss": 0.2, "val_loss": 0.1}


def test_convert_to_metrics_with_list_of_floats():
    assert tuner_utils.convert_to_metrics_dict(
        [0.1, 0.2],
        obj_module.Objective("val_loss", "min"),
    ) == {"val_loss": (0.1 + 0.2) / 2}


def test_convert_to_metrics_with_dict_without_obj_key():
    with pytest.raises(ValueError, match="the specified objective"):
        tuner_utils.validate_trial_results(
            {"loss": 0.1}, obj_module.Objective("val_loss", "min"), "func_name"
        )


def test_get_best_step_return_zero():
    assert (
        tuner_utils.get_best_step(
            [{"val_loss": 1}, {"val_loss": 2}],
            obj_module.Objective("val_loss", "min"),
        )
        == 0
    )


def test_get_best_step_return_average_epoch():
    class History(keras.callbacks.History):
        def __init__(self, history):
            self.history = history

    results = [
        History(
            {
                "val_loss": [5, 8, 3, 1, 2],
                "val_accuracy": [5, 8, 3, 1, 2],
            }
        ),
        History(
            {
                "val_loss": [5, 8, 3, 2, 1],
                "val_accuracy": [5, 8, 3, 1, 2],
            }
        ),
    ]
    assert (
        tuner_utils.get_best_step(
            results,
            obj_module.Objective("val_loss", "min"),
        )
        == 3
    )


def test_display_format_duration_large_d():
    oracle = gridsearch.GridSearchOracle()
    d = datetime.datetime(2020, 5, 17) - datetime.datetime(2020, 5, 10)
    assert tuner_utils.Display(oracle).format_duration(d) == "7d 00h 00m 00s"



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_tuner.engine import hyperparameters as hp_module


def test_base_hyperparameter():
    base_param = hp_module.HyperParameter(name="base", default=0)
    assert base_param.name == "base"
    assert base_param.default == 0
    assert base_param.get_config() == {
        "name": "base",
        "default": 0,
        "conditions": [],
    }
    base_param = hp_module.HyperParameter.from_config(base_param.get_config())
    assert base_param.name == "base"
    assert base_param.default == 0



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle

import numpy as np
import pytest
from sklearn import linear_model

import keras_tuner
from keras_tuner import errors
from keras_tuner.engine import base_tuner
from keras_tuner.tuners import gridsearch
from keras_tuner.tuners import randomsearch

INPUT_DIM = 2
NUM_CLASSES = 3
NUM_SAMPLES = 64
TRAIN_INPUTS = np.random.random(size=(NUM_SAMPLES, INPUT_DIM))
TRAIN_TARGETS = np.random.randint(0, NUM_CLASSES, size=(NUM_SAMPLES,))
VAL_INPUTS = np.random.random(size=(NUM_SAMPLES, INPUT_DIM))
VAL_TARGETS = np.random.randint(0, NUM_CLASSES, size=(NUM_SAMPLES,))


def test_base_tuner(tmp_path):
    class MyTuner(base_tuner.BaseTuner):
        def run_trial(self, trial, x):
            model = self.hypermodel.build(trial.hyperparameters)
            self.oracle.update_space(trial.hyperparameters)
            score = model(x)
            return {"score": score}

        def get_best_models(self, num_models=1):
            best_trials = self.oracle.get_best_trials(num_models)
            models = [
                self.hypermodel.build(t.hyperparameters) for t in best_trials
            ]
            return models

    def build_model(hp):
        class MyModel:
            def __init__(self):
                self.factor = hp.Float("a", 0, 10)

            def __call__(self, x):
                return self.factor * x

        return MyModel()

    oracle = randomsearch.RandomSearchOracle(
        objective=keras_tuner.Objective("score", "max"), max_trials=5
    )
    tuner = MyTuner(oracle=oracle, hypermodel=build_model, directory=tmp_path)
    tuner.search(1.0)
    models = tuner.get_best_models(5)

    # Check that scoring of the model was done correctly.
    models_by_factor = sorted(models, key=lambda m: m.factor, reverse=True)
    assert models[0] == models_by_factor[0]


def test_oracle_idle(tmp_path):
    class MyOracle(randomsearch.RandomSearchOracle):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._first_trial = True

        def create_trial(self, *args, **kwargs):
            trial = super().create_trial(*args, **kwargs)
            if self._first_trial:
                self._first_trial = False
                trial.status = "IDLE"
            else:
                trial.status = "STOPPED"
            return trial

    class MyTuner(base_tuner.BaseTuner):
        def run_trial(self, trial, *args, **kwargs):
            trial.hyperparameters.Boolean("foo")
            trial.hyperparameters.Boolean("bar")
            return np.random.rand()

    tuner = MyTuner(oracle=MyOracle(max_trials=3), directory=tmp_path)
    tuner.search()


def test_simple_sklearn_tuner(tmp_path):
    class SimpleSklearnTuner(base_tuner.BaseTuner):
        def run_trial(self, trial, x, y, validation_data):
            model = self.hypermodel.build(trial.hyperparameters)
            model.fit(x, y)
            x_val, y_val = validation_data
            score = model.score(x_val, y_val)
            self.save_model(trial.trial_id, model)
            return {"score": score}

        def save_model(self, trial_id, model, step=0):
            fname = os.path.join(self.get_trial_dir(trial_id), "model.pickle")
            with open(fname, "wb") as f:
                pickle.dump(model, f)

        def load_model(self, trial):
            fname = os.path.join(
                self.get_trial_dir(trial.trial_id), "model.pickle"
            )
            with open(fname, "rb") as f:
                return pickle.load(f)

    def sklearn_build_fn(hp):
        c = hp.Float("c", 1e-4, 10)
        return linear_model.LogisticRegression(C=c)

    tuner = SimpleSklearnTuner(
        oracle=randomsearch.RandomSearchOracle(
            objective=keras_tuner.Objective("score", "max"), max_trials=2
        ),
        hypermodel=sklearn_build_fn,
        directory=tmp_path,
    )
    tuner.search(
        TRAIN_INPUTS, TRAIN_TARGETS, validation_data=(VAL_INPUTS, VAL_TARGETS)
    )
    models = tuner.get_best_models(2)
    score0 = models[0].score(VAL_INPUTS, VAL_TARGETS)
    score1 = models[1].score(VAL_INPUTS, VAL_TARGETS)
    assert score0 >= score1


def test_retry_common_errors(tmp_path):
    class MyTuner(gridsearch.GridSearch):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.count = 0

        def run_trial(self, trial, *fit_args, **fit_kwargs):
            self.count += 1
            hp = trial.hyperparameters
            value = hp.Int("value", min_value=1, max_value=5)
            if value in [2, 3]:
                raise ValueError(f"wrong value {value}.")
            return value

    tuner = MyTuner(directory=tmp_path, max_retries_per_trial=2)
    tuner.search()
    assert tuner.count == 5 + 2 * 2  # 5 values, 2 values * 2 retries


def test_fatal_error_is_raise_again(tmp_path):
    class MyTuner(gridsearch.GridSearch):
        def run_trial(self, trial, *fit_args, **fit_kwargs):
            raise errors.FatalError("FatalError raised.")

    tuner = MyTuner(directory=tmp_path, max_retries_per_trial=2)
    with pytest.raises(errors.FatalError, match="FatalError raised."):
        tuner.search()


def test_failed_trial_error_no_retry(tmp_path):
    class MyTuner(gridsearch.GridSearch):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.count = 0

        def run_trial(self, trial, *fit_args, **fit_kwargs):
            self.count += 1
            hp = trial.hyperparameters
            value = hp.Int("value", min_value=1, max_value=5)
            if value in [2, 3]:
                raise errors.FailedTrialError(f"wrong value {value}.")
            return value

    tuner = MyTuner(directory=tmp_path, max_retries_per_trial=2)
    tuner.search()
    assert tuner.count == 5  # 5 values, no retry


def test_remaining_trials(tmp_path):
    class MyTuner(gridsearch.GridSearch):
        def run_trial(self, trial, *fit_args, **fit_kwargs):
            raise errors.FatalError("FatalError raised.")

    tuner = MyTuner(directory=tmp_path, max_retries_per_trial=2, max_trials=200)
    assert tuner.remaining_trials == 200


def test_logger_deprecated(tmp_path):
    def build_model(hp):
        hp.Boolean("a")

    with pytest.deprecated_call(match="logger"):
        gridsearch.GridSearch(
            directory=tmp_path,
            hypermodel=build_model,
            max_trials=200,
            logger=1,
        )


def test_unrecognized_arguments_raise_value_error(tmp_path):
    def build_model(hp):
        hp.Boolean("a")

    with pytest.raises(ValueError, match="Unrecognized arguments"):
        gridsearch.GridSearch(
            directory=tmp_path,
            hypermodel=build_model,
            max_trials=200,
            unrecognized=3,
        )



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from keras_tuner import protos


def sampling_from_proto(sampling):
    if sampling == protos.get_proto().Sampling.LINEAR:
        return "linear"
    if sampling == protos.get_proto().Sampling.LOG:
        return "log"
    if sampling == protos.get_proto().Sampling.REVERSE_LOG:
        return "reverse_log"
    raise ValueError(
        "Expected sampling to be one of predefined proto values. "
        f"Received: '{sampling}'."
    )


def sampling_to_proto(sampling):
    if sampling == "linear":
        return protos.get_proto().Sampling.LINEAR
    if sampling == "log":
        return protos.get_proto().Sampling.LOG
    if sampling == "reverse_log":
        return protos.get_proto().Sampling.REVERSE_LOG
    raise ValueError(
        "Expected sampling to be 'linear', 'log', or 'reverse_log'. "
        f"Received: '{sampling}'."
    )


def prob_to_index(prob, n_index):
    """Convert cumulative probability to 0-based index in the given range."""
    ele_prob = 1 / n_index
    index = int(math.floor(prob / ele_prob))
    # Can happen when `prob` is very close to 1.
    if index == n_index:
        index -= 1
    return index


def index_to_prob(index, n_index):
    """Convert 0-based index in the given range to cumulative probability."""
    ele_prob = 1 / n_index
    # Center the value in its probability bucket.
    return (index + 0.5) * ele_prob



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import tensorflow as tf

from keras_tuner.engine import tuner as tuner_module
from keras_tuner.tuners import randomsearch


def test_update_space(tmp_path):
    # Tests that HyperParameters added after the first call to `build_model`
    # are sent to the Oracle via oracle.update_space.
    def build_model(hp):
        model = tf.keras.Sequential()
        for i in range(hp.Int("layers", 0, 2)):
            model.add(
                tf.keras.layers.Dense(
                    units=hp.Int(f"units_{str(i)}", 2, 4, 2), activation="relu"
                )
            )

        model.add(tf.keras.layers.Dense(1, activation="sigmoid"))
        model.compile("adam", loss="binary_crossentropy", metrics=["accuracy"])
        return model

    class MyRandomSearch(randomsearch.RandomSearchOracle):
        def populate_space(self, trial_id):
            result = super().populate_space(trial_id)
            if "values" in result:
                result["values"]["layers"] = 2
            return result

    tuner = tuner_module.Tuner(
        oracle=MyRandomSearch(objective="accuracy", max_trials=1),
        hypermodel=build_model,
        directory=tmp_path,
    )

    assert {hp.name for hp in tuner.oracle.get_space().space} == {"layers"}

    x, y = np.ones((10, 10)), np.ones((10, 1))
    tuner.search(x, y, epochs=1)

    assert {hp.name for hp in tuner.oracle.get_space().space} == {
        "layers",
        "units_0",
        "units_1",
    }


def test_exhausted_search_space(tmp_path):
    class MyTuner(randomsearch.RandomSearch):
        def run_trial(self, trial, *args, **kwargs):
            hp = trial.hyperparameters
            hp.Boolean("boolean")
            hp.Boolean("boolean2")
            return [np.random.rand()]

    tuner = MyTuner(
        max_trials=15,
        directory=tmp_path,
    )
    tuner.search()
    assert len(tuner.oracle.trials) == 4



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

try:
    import sklearn
    import sklearn.exceptions
    import sklearn.gaussian_process
except ImportError:  # pragma: no cover
    sklearn = None  # pragma: no cover

try:
    import scipy
    import scipy.optimize
except ImportError:  # pragma: no cover
    scipy = None  # pragma: no cover

from keras_tuner.api_export import keras_tuner_export
from keras_tuner.engine import hyperparameters as hp_module
from keras_tuner.engine import oracle as oracle_module
from keras_tuner.engine import trial as trial_module
from keras_tuner.engine import tuner as tuner_module


@keras_tuner_export("keras_tuner.oracles.BayesianOptimizationOracle")
class BayesianOptimizationOracle(oracle_module.Oracle):
    """Bayesian optimization oracle.

    It uses Bayesian optimization with a underlying Gaussian process model.
    The acquisition function used is upper confidence bound (UCB), which can
    be found [here](
    https://www.cse.wustl.edu/~garnett/cse515t/spring_2015/files/lecture_notes/12.pdf).

    Args:
        objective: A string, `keras_tuner.Objective` instance, or a list of
            `keras_tuner.Objective`s and strings. If a string, the direction of
            the optimization (min or max) will be inferred. If a list of
            `keras_tuner.Objective`, we will minimize the sum of all the
            objectives to minimize subtracting the sum of all the objectives to
            maximize. The `objective` argument is optional when
            `Tuner.run_trial()` or `HyperModel.fit()` returns a single float as
            the objective to minimize.
        max_trials: Integer, the total number of trials (model configurations)
            to test at most. Note that the oracle may interrupt the search
            before `max_trial` models have been tested if the search space has
            been exhausted. Defaults to 10.
        num_initial_points: Optional number of randomly generated samples as
            initial training data for Bayesian optimization. If left
            unspecified, a value of 3 times the dimensionality of the
            hyperparameter space is used.
        alpha: Float, the value added to the diagonal of the kernel matrix
            during fitting. It represents the expected amount of noise in the
            observed performances in Bayesian optimization. Defaults to 1e-4.
        beta: Float, the balancing factor of exploration and exploitation. The
            larger it is, the more explorative it is. Defaults to 2.6.
        seed: Optional integer, the random seed.
        hyperparameters: Optional `HyperParameters` instance. Can be used to
            override (or register in advance) hyperparameters in the search
            space.
        tune_new_entries: Boolean, whether hyperparameter entries that are
            requested by the hypermodel but that were not specified in
            `hyperparameters` should be added to the search space, or not. If
            not, then the default value for these parameters will be used.
            Defaults to True.
        allow_new_entries: Boolean, whether the hypermodel is allowed to
            request hyperparameter entries not listed in `hyperparameters`.
            Defaults to True.
        max_retries_per_trial: Integer. Defaults to 0. The maximum number of
            times to retry a `Trial` if the trial crashed or the results are
            invalid.
        max_consecutive_failed_trials: Integer. Defaults to 3. The maximum
            number of consecutive failed `Trial`s. When this number is reached,
            the search will be stopped. A `Trial` is marked as failed when none
            of the retries succeeded.
    """

    def __init__(
        self,
        objective=None,
        max_trials=10,
        num_initial_points=None,
        alpha=1e-4,
        beta=2.6,
        seed=None,
        hyperparameters=None,
        allow_new_entries=True,
        tune_new_entries=True,
        max_retries_per_trial=0,
        max_consecutive_failed_trials=3,
    ):
        if scipy is None:
            raise ImportError(
                "Please install scipy before using the `BayesianOptimization` "
                "with `pip install keras-tuner[bayesian]`."
            )

        if sklearn is None:
            raise ImportError(
                "Please install scikit-learn (sklearn) before using the "
                "`BayesianOptimization` with "
                "`pip install keras-tuner[bayesian]`."
            )
        super().__init__(
            objective=objective,
            max_trials=max_trials,
            hyperparameters=hyperparameters,
            tune_new_entries=tune_new_entries,
            allow_new_entries=allow_new_entries,
            seed=seed,
            max_retries_per_trial=max_retries_per_trial,
            max_consecutive_failed_trials=max_consecutive_failed_trials,
        )
        self.num_initial_points = num_initial_points
        self.alpha = alpha
        self.beta = beta
        self._random_state = np.random.RandomState(self.seed)
        self.gpr = self._make_gpr()

    def _make_gpr(self):
        return sklearn.gaussian_process.GaussianProcessRegressor(
            kernel=sklearn.gaussian_process.kernels.Matern(nu=2.5),
            n_restarts_optimizer=20,
            normalize_y=True,
            alpha=self.alpha,
            random_state=self.seed,
        )

    def populate_space(self, trial_id):
        """Fill the hyperparameter space with values.

        Args:
            trial_id: A string, the ID for this Trial.

        Returns:
            A dictionary with keys "values" and "status", where "values" is
            a mapping of parameter names to suggested values, and "status"
            should be one of "RUNNING" (the trial can start normally), "IDLE"
            (the oracle is waiting on something and cannot create a trial), or
            "STOPPED" (the oracle has finished searching and no new trial should
            be created).
        """
        # Generate enough samples before training Gaussian process.
        completed_trials = [
            t for t in self.trials.values() if t.status == "COMPLETED"
        ]

        # Use 3 times the dimensionality of the space as the default number of
        # random points.
        dimensions = len(self.hyperparameters.space)
        num_initial_points = self.num_initial_points or max(3 * dimensions, 3)
        if len(completed_trials) < num_initial_points:
            return self._random_populate_space()

        # Fit a GPR to the completed trials and return the predicted optimum
        # values.
        x, y = self._vectorize_trials()

        # Ensure no nan, inf in x, y. GPR cannot process nan or inf.
        x = np.nan_to_num(x, posinf=0, neginf=0)
        y = np.nan_to_num(y, posinf=0, neginf=0)

        self.gpr.fit(x, y)

        def _upper_confidence_bound(x):
            x = x.reshape(1, -1)
            mu, sigma = self.gpr.predict(x, return_std=True)
            return mu - self.beta * sigma

        optimal_val = float("inf")
        optimal_x = None
        num_restarts = 50
        bounds = self._get_hp_bounds()
        x_seeds = self._random_state.uniform(
            bounds[:, 0], bounds[:, 1], size=(num_restarts, bounds.shape[0])
        )
        for x_try in x_seeds:
            # Sign of score is flipped when maximizing.
            result = scipy.optimize.minimize(
                _upper_confidence_bound,
                x0=x_try,
                bounds=bounds,
                method="L-BFGS-B",
            )
            result_fun = (
                result.fun if np.isscalar(result.fun) else result.fun[0]
            )
            if result_fun < optimal_val:
                optimal_val = result_fun
                optimal_x = result.x

        values = self._vector_to_values(optimal_x)
        return {"status": trial_module.TrialStatus.RUNNING, "values": values}

    def _random_populate_space(self):
        values = self._random_values()
        if values is None:
            return {"status": trial_module.TrialStatus.STOPPED, "values": None}
        return {"status": trial_module.TrialStatus.RUNNING, "values": values}

    def get_state(self):
        state = super().get_state()
        state.update(
            {
                "num_initial_points": self.num_initial_points,
                "alpha": self.alpha,
                "beta": self.beta,
            }
        )
        return state

    def set_state(self, state):
        super().set_state(state)
        self.num_initial_points = state["num_initial_points"]
        self.alpha = state["alpha"]
        self.beta = state["beta"]
        self.gpr = self._make_gpr()

    def _vectorize_trials(self):
        x = []
        y = []
        ongoing_trials = set(self.ongoing_trials.values())
        for trial in self.trials.values():
            # Create a vector representation of each Trial's hyperparameters.
            trial_hps = trial.hyperparameters
            vector = []
            for hp in self._nonfixed_space():
                # For hyperparameters not present in the trial (either added
                # after the trial or inactive in the trial), set to default
                # value.
                if (
                    trial_hps.is_active(hp)  # inactive
                    and hp.name in trial_hps.values  # added after the trial
                ):
                    trial_value = trial_hps.values[hp.name]
                else:
                    trial_value = hp.default

                # Embed an HP value into the continuous space [0, 1].
                prob = hp.value_to_prob(trial_value)
                vector.append(prob)

            if trial in ongoing_trials:
                # "Hallucinate" the results of ongoing trials. This ensures that
                # repeat trials are not selected when running distributed.
                x_h = np.array(vector).reshape((1, -1))
                y_h_mean, y_h_std = self.gpr.predict(x_h, return_std=True)
                # Give a pessimistic estimate of the ongoing trial.
                score = y_h_mean[0] + y_h_std[0]
            elif trial.status == "COMPLETED":
                score = trial.score
                # Always frame the optimization as a minimization for
                # scipy.minimize.
                if self.objective.direction == "max":
                    score = -1 * score
            elif trial.status in ["FAILED", "INVALID"]:
                # Skip the failed and invalid trials.
                continue

            x.append(vector)
            y.append(score)

        x = np.array(x)
        y = np.array(y)
        return x, y

    def _vector_to_values(self, vector):
        hps = hp_module.HyperParameters()
        vector_index = 0
        for hp in self.hyperparameters.space:
            hps.merge([hp])
            if isinstance(hp, hp_module.Fixed):
                value = hp.value
            else:
                prob = vector[vector_index]
                vector_index += 1
                value = hp.prob_to_value(prob)

            if hps.is_active(hp):
                hps.values[hp.name] = value
        return hps.values

    def _nonfixed_space(self):
        return [
            hp
            for hp in self.hyperparameters.space
            if not isinstance(hp, hp_module.Fixed)
        ]

    def _get_hp_bounds(self):
        bounds = [[0, 1] for _ in self._nonfixed_space()]
        return np.array(bounds)


@keras_tuner_export(
    [
        "keras_tuner.BayesianOptimization",
        "keras_tuner.tuners.BayesianOptimization",
    ]
)
class BayesianOptimization(tuner_module.Tuner):
    """BayesianOptimization tuning with Gaussian process.

    Args:
        hypermodel: Instance of `HyperModel` class (or callable that takes
            hyperparameters and returns a `Model` instance). It is optional
            when `Tuner.run_trial()` is overriden and does not use
            `self.hypermodel`.
        objective: A string, `keras_tuner.Objective` instance, or a list of
            `keras_tuner.Objective`s and strings. If a string, the direction of
            the optimization (min or max) will be inferred. If a list of
            `keras_tuner.Objective`, we will minimize the sum of all the
            objectives to minimize subtracting the sum of all the objectives to
            maximize. The `objective` argument is optional when
            `Tuner.run_trial()` or `HyperModel.fit()` returns a single float as
            the objective to minimize.
        max_trials: Integer, the total number of trials (model configurations)
            to test at most. Note that the oracle may interrupt the search
            before `max_trial` models have been tested if the search space has
            been exhausted. Defaults to 10.
        num_initial_points: Optional number of randomly generated samples as
            initial training data for Bayesian optimization. If left
            unspecified, a value of 3 times the dimensionality of the
            hyperparameter space is used.
        alpha: Float, the value added to the diagonal of the kernel matrix
            during fitting. It represents the expected amount of noise in the
            observed performances in Bayesian optimization. Defaults to 1e-4.
        beta: Float, the balancing factor of exploration and exploitation. The
            larger it is, the more explorative it is. Defaults to 2.6.
        seed: Optional integer, the random seed.
        hyperparameters: Optional `HyperParameters` instance. Can be used to
            override (or register in advance) hyperparameters in the search
            space.
        tune_new_entries: Boolean, whether hyperparameter entries that are
            requested by the hypermodel but that were not specified in
            `hyperparameters` should be added to the search space, or not. If
            not, then the default value for these parameters will be used.
            Defaults to True.
        allow_new_entries: Boolean, whether the hypermodel is allowed to
            request hyperparameter entries not listed in `hyperparameters`.
            Defaults to True.
        max_retries_per_trial: Integer. Defaults to 0. The maximum number of
            times to retry a `Trial` if the trial crashed or the results are
            invalid.
        max_consecutive_failed_trials: Integer. Defaults to 3. The maximum
            number of consecutive failed `Trial`s. When this number is reached,
            the search will be stopped. A `Trial` is marked as failed when none
            of the retries succeeded.
        **kwargs: Keyword arguments relevant to all `Tuner` subclasses. Please
            see the docstring for `Tuner`.
    """

    def __init__(
        self,
        hypermodel=None,
        objective=None,
        max_trials=10,
        num_initial_points=None,
        alpha=1e-4,
        beta=2.6,
        seed=None,
        hyperparameters=None,
        tune_new_entries=True,
        allow_new_entries=True,
        max_retries_per_trial=0,
        max_consecutive_failed_trials=3,
        **kwargs
    ):
        oracle = BayesianOptimizationOracle(
            objective=objective,
            max_trials=max_trials,
            num_initial_points=num_initial_points,
            alpha=alpha,
            beta=beta,
            seed=seed,
            hyperparameters=hyperparameters,
            tune_new_entries=tune_new_entries,
            allow_new_entries=allow_new_entries,
            max_retries_per_trial=max_retries_per_trial,
            max_consecutive_failed_trials=max_consecutive_failed_trials,
        )
        super().__init__(oracle=oracle, hypermodel=hypermodel, **kwargs)



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from keras_tuner.protos.v3 import (
    service_pb2 as keras__tuner_dot_protos_dot_service__pb2,
)


class OracleStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetSpace = channel.unary_unary(
            "/keras_tuner.Oracle/GetSpace",
            request_serializer=keras__tuner_dot_protos_dot_service__pb2.GetSpaceRequest.SerializeToString,
            response_deserializer=keras__tuner_dot_protos_dot_service__pb2.GetSpaceResponse.FromString,
        )
        self.UpdateSpace = channel.unary_unary(
            "/keras_tuner.Oracle/UpdateSpace",
            request_serializer=keras__tuner_dot_protos_dot_service__pb2.UpdateSpaceRequest.SerializeToString,
            response_deserializer=keras__tuner_dot_protos_dot_service__pb2.UpdateSpaceResponse.FromString,
        )
        self.CreateTrial = channel.unary_unary(
            "/keras_tuner.Oracle/CreateTrial",
            request_serializer=keras__tuner_dot_protos_dot_service__pb2.CreateTrialRequest.SerializeToString,
            response_deserializer=keras__tuner_dot_protos_dot_service__pb2.CreateTrialResponse.FromString,
        )
        self.UpdateTrial = channel.unary_unary(
            "/keras_tuner.Oracle/UpdateTrial",
            request_serializer=keras__tuner_dot_protos_dot_service__pb2.UpdateTrialRequest.SerializeToString,
            response_deserializer=keras__tuner_dot_protos_dot_service__pb2.UpdateTrialResponse.FromString,
        )
        self.EndTrial = channel.unary_unary(
            "/keras_tuner.Oracle/EndTrial",
            request_serializer=keras__tuner_dot_protos_dot_service__pb2.EndTrialRequest.SerializeToString,
            response_deserializer=keras__tuner_dot_protos_dot_service__pb2.EndTrialResponse.FromString,
        )
        self.GetBestTrials = channel.unary_unary(
            "/keras_tuner.Oracle/GetBestTrials",
            request_serializer=keras__tuner_dot_protos_dot_service__pb2.GetBestTrialsRequest.SerializeToString,
            response_deserializer=keras__tuner_dot_protos_dot_service__pb2.GetBestTrialsResponse.FromString,
        )
        self.GetTrial = channel.unary_unary(
            "/keras_tuner.Oracle/GetTrial",
            request_serializer=keras__tuner_dot_protos_dot_service__pb2.GetTrialRequest.SerializeToString,
            response_deserializer=keras__tuner_dot_protos_dot_service__pb2.GetTrialResponse.FromString,
        )


class OracleServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetSpace(self, request, context):
        """Return the HyperParameter search space."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def UpdateSpace(self, request, context):
        """Updates the HyperParameter search space."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CreateTrial(self, request, context):
        """Creates a Trial."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def UpdateTrial(self, request, context):
        """Updates a Trial with metrics and a step."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def EndTrial(self, request, context):
        """Ends a Trial."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetBestTrials(self, request, context):
        """Gets the best Trials."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetTrial(self, request, context):
        """Gets a Trial by ID."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_OracleServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "GetSpace": grpc.unary_unary_rpc_method_handler(
            servicer.GetSpace,
            request_deserializer=keras__tuner_dot_protos_dot_service__pb2.GetSpaceRequest.FromString,
            response_serializer=keras__tuner_dot_protos_dot_service__pb2.GetSpaceResponse.SerializeToString,
        ),
        "UpdateSpace": grpc.unary_unary_rpc_method_handler(
            servicer.UpdateSpace,
            request_deserializer=keras__tuner_dot_protos_dot_service__pb2.UpdateSpaceRequest.FromString,
            response_serializer=keras__tuner_dot_protos_dot_service__pb2.UpdateSpaceResponse.SerializeToString,
        ),
        "CreateTrial": grpc.unary_unary_rpc_method_handler(
            servicer.CreateTrial,
            request_deserializer=keras__tuner_dot_protos_dot_service__pb2.CreateTrialRequest.FromString,
            response_serializer=keras__tuner_dot_protos_dot_service__pb2.CreateTrialResponse.SerializeToString,
        ),
        "UpdateTrial": grpc.unary_unary_rpc_method_handler(
            servicer.UpdateTrial,
            request_deserializer=keras__tuner_dot_protos_dot_service__pb2.UpdateTrialRequest.FromString,
            response_serializer=keras__tuner_dot_protos_dot_service__pb2.UpdateTrialResponse.SerializeToString,
        ),
        "EndTrial": grpc.unary_unary_rpc_method_handler(
            servicer.EndTrial,
            request_deserializer=keras__tuner_dot_protos_dot_service__pb2.EndTrialRequest.FromString,
            response_serializer=keras__tuner_dot_protos_dot_service__pb2.EndTrialResponse.SerializeToString,
        ),
        "GetBestTrials": grpc.unary_unary_rpc_method_handler(
            servicer.GetBestTrials,
            request_deserializer=keras__tuner_dot_protos_dot_service__pb2.GetBestTrialsRequest.FromString,
            response_serializer=keras__tuner_dot_protos_dot_service__pb2.GetBestTrialsResponse.SerializeToString,
        ),
        "GetTrial": grpc.unary_unary_rpc_method_handler(
            servicer.GetTrial,
            request_deserializer=keras__tuner_dot_protos_dot_service__pb2.GetTrialRequest.FromString,
            response_serializer=keras__tuner_dot_protos_dot_service__pb2.GetTrialResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "keras_tuner.Oracle", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class Oracle(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetSpace(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/keras_tuner.Oracle/GetSpace",
            keras__tuner_dot_protos_dot_service__pb2.GetSpaceRequest.SerializeToString,
            keras__tuner_dot_protos_dot_service__pb2.GetSpaceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def UpdateSpace(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/keras_tuner.Oracle/UpdateSpace",
            keras__tuner_dot_protos_dot_service__pb2.UpdateSpaceRequest.SerializeToString,
            keras__tuner_dot_protos_dot_service__pb2.UpdateSpaceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CreateTrial(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/keras_tuner.Oracle/CreateTrial",
            keras__tuner_dot_protos_dot_service__pb2.CreateTrialRequest.SerializeToString,
            keras__tuner_dot_protos_dot_service__pb2.CreateTrialResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def UpdateTrial(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/keras_tuner.Oracle/UpdateTrial",
            keras__tuner_dot_protos_dot_service__pb2.UpdateTrialRequest.SerializeToString,
            keras__tuner_dot_protos_dot_service__pb2.UpdateTrialResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def EndTrial(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/keras_tuner.Oracle/EndTrial",
            keras__tuner_dot_protos_dot_service__pb2.EndTrialRequest.SerializeToString,
            keras__tuner_dot_protos_dot_service__pb2.EndTrialResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetBestTrials(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/keras_tuner.Oracle/GetBestTrials",
            keras__tuner_dot_protos_dot_service__pb2.GetBestTrialsRequest.SerializeToString,
            keras__tuner_dot_protos_dot_service__pb2.GetBestTrialsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetTrial(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/keras_tuner.Oracle/GetTrial",
            keras__tuner_dot_protos_dot_service__pb2.GetTrialRequest.SerializeToString,
            keras__tuner_dot_protos_dot_service__pb2.GetTrialResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for distributed tuning."""

import copy
import logging
import os
import threading
from unittest import mock

import numpy as np
import portpicker
import pytest
import tensorflow as tf
from tensorflow import keras

import keras_tuner
from keras_tuner.distribute import oracle_client
from keras_tuner.distribute import utils as dist_utils
from keras_tuner.engine import hyperparameters as hp_module
from keras_tuner.test_utils import mock_distribute
from keras_tuner.tuners import randomsearch


class SimpleTuner(keras_tuner.engine.base_tuner.BaseTuner):
    def run_trial(self, trial):
        score = self.hypermodel.build(trial.hyperparameters)
        self.save_model(trial.trial_id, score)
        return {"score": score}

    def save_model(self, trial_id, score, step=0):
        save_path = os.path.join(self.project_dir, trial_id)
        with tf.io.gfile.GFile(save_path, "w") as f:
            f.write(str(score))

    def load_model(self, trial):
        save_path = os.path.join(self.project_dir, trial.trial_id)
        with tf.io.gfile.GFile(save_path, "r") as f:
            score = int(f.read())
        return score


def test_base_tuner_distribution(tmp_path):
    num_workers = 3
    barrier = threading.Barrier(num_workers)

    def _test_base_tuner():
        def build_model(hp):
            return hp.Int("a", 1, 100)

        tuner = SimpleTuner(
            oracle=keras_tuner.oracles.RandomSearchOracle(
                objective=keras_tuner.Objective("score", "max"), max_trials=10
            ),
            hypermodel=build_model,
            directory=tmp_path,
        )
        tuner.search()

        # Only worker makes it to this point, server runs until thread stops.
        assert dist_utils.has_chief_oracle()
        assert not dist_utils.is_chief_oracle()
        assert isinstance(
            tuner.oracle, keras_tuner.distribute.oracle_client.OracleClient
        )

        barrier.wait(60)

        # Model is just a score.
        scores = tuner.get_best_models(10)
        assert len(scores)
        assert scores == sorted(copy.copy(scores), reverse=True)

    mock_distribute.mock_distribute(_test_base_tuner, num_workers=num_workers)


def test_random_search(tmp_path):
    # TensorFlow model building and execution is not thread-safe.
    num_workers = 1

    def _test_random_search():
        def build_model(hp):
            model = keras.Sequential()
            model.add(keras.layers.Dense(3, input_shape=(5,)))
            for i in range(hp.Int("num_layers", 1, 3)):
                model.add(
                    keras.layers.Dense(
                        hp.Int("num_units_%i" % i, 1, 3), activation="relu"
                    )
                )
            model.add(keras.layers.Dense(1, activation="sigmoid"))
            model.compile("sgd", "binary_crossentropy")
            return model

        x = np.random.uniform(-1, 1, size=(2, 5))
        y = np.ones((2, 1))

        tuner = keras_tuner.tuners.RandomSearch(
            hypermodel=build_model,
            objective="val_loss",
            max_trials=10,
            directory=tmp_path,
        )

        # Only worker makes it to this point, server runs until thread stops.
        assert dist_utils.has_chief_oracle()
        assert not dist_utils.is_chief_oracle()
        assert isinstance(
            tuner.oracle, keras_tuner.distribute.oracle_client.OracleClient
        )

        tuner.search(x, y, validation_data=(x, y), epochs=1, batch_size=2)

        # Suppress warnings about optimizer state not being restored by
        # tf.keras.
        tf.get_logger().setLevel(logging.ERROR)

        trials = tuner.oracle.get_best_trials(2)
        assert trials[0].score <= trials[1].score

        models = tuner.get_best_models(2)
        assert models[0].evaluate(x, y) <= models[1].evaluate(x, y)

    mock_distribute.mock_distribute(_test_random_search, num_workers)


def test_client_no_attribute_error():
    with mock.patch.object(os, "environ", mock_distribute.MockEnvVars()):
        port = str(portpicker.pick_unused_port())
        os.environ["KERASTUNER_ORACLE_IP"] = "127.0.0.1"
        os.environ["KERASTUNER_ORACLE_PORT"] = port
        os.environ["KERASTUNER_TUNER_ID"] = "worker0"
        hps = keras_tuner.HyperParameters()
        oracle = randomsearch.RandomSearchOracle(
            objective=keras_tuner.Objective("score", "max"),
            max_trials=10,
            hyperparameters=hps,
        )
        client = oracle_client.OracleClient(oracle)
        with pytest.raises(AttributeError, match="has no attribute"):
            client.unknown_attribute


@mock.patch("keras_tuner.distribute.oracle_client.OracleClient.get_space")
def test_should_not_report_update_trial_return_running(get_space):
    get_space.return_value = hp_module.HyperParameters()
    with mock.patch.object(os, "environ", mock_distribute.MockEnvVars()):
        port = str(portpicker.pick_unused_port())
        os.environ["KERASTUNER_ORACLE_IP"] = "127.0.0.1"
        os.environ["KERASTUNER_ORACLE_PORT"] = port
        os.environ["KERASTUNER_TUNER_ID"] = "worker0"
        hps = keras_tuner.HyperParameters()
        oracle = randomsearch.RandomSearchOracle(
            objective=keras_tuner.Objective("score", "max"),
            max_trials=10,
            hyperparameters=hps,
        )
        client = oracle_client.OracleClient(oracle)
        client.should_report = False
        assert client.update_trial("a", {"score": 100}).status == "RUNNING"



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from keras_tuner.engine import hyperparameters as hp_module


def test_boolean():
    # Test default default
    boolean = hp_module.Boolean("bool")
    assert boolean.default is False
    # Test default setting
    boolean = hp_module.Boolean("bool", default=True)
    assert boolean.default is True
    # Wrong default type
    with pytest.raises(ValueError, match="must be a Python boolean"):
        hp_module.Boolean("bool", default=None)
    # Test serialization
    boolean = hp_module.Boolean("bool", default=True)
    boolean = hp_module.Boolean.from_config(boolean.get_config())
    assert boolean.default is True
    assert boolean.name == "bool"

    # Test random_sample
    assert boolean.random_sample() in {True, False}
    assert boolean.random_sample(123) == boolean.random_sample(123)
    assert {boolean.value_to_prob(True), boolean.value_to_prob(False)} == {
        0.25,
        0.75,
    }


def test_boolean_repr():
    assert repr(hp_module.Boolean("bool")) == repr(hp_module.Boolean("bool"))


def test_boolean_values_property():
    assert list(hp_module.Boolean("bool").values) == [True, False]



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for Tuner class."""


import collections
import statistics
from datetime import datetime

import numpy as np
from tensorboard.plugins.hparams import api as hparams_api
from tensorflow import keras

from keras_tuner import errors
from keras_tuner import utils
from keras_tuner.distribute import file_utils
from keras_tuner.engine import hyperparameters as hp_module
from keras_tuner.engine import objective as obj_module


class TunerCallback(keras.callbacks.Callback):
    def __init__(self, tuner, trial):
        super().__init__()
        self.tuner = tuner
        self.trial = trial

    def on_epoch_begin(self, epoch, logs=None):
        self.tuner.on_epoch_begin(self.trial, self.model, epoch, logs=logs)

    def on_batch_begin(self, batch, logs=None):
        self.tuner.on_batch_begin(self.trial, self.model, batch, logs)

    def on_batch_end(self, batch, logs=None):
        self.tuner.on_batch_end(self.trial, self.model, batch, logs)

    def on_epoch_end(self, epoch, logs=None):
        self.tuner.on_epoch_end(self.trial, self.model, epoch, logs=logs)


# TODO: Add more extensive display.
class Display:
    def __init__(self, oracle, verbose=1):
        self.verbose = verbose
        self.oracle = oracle
        self.col_width = 18

        # Start time for the overall search
        self.search_start = None

        # Start time of the latest trial
        self.trial_start = None

    def on_trial_begin(self, trial):
        if self.verbose >= 1:
            self.trial_number = int(trial.trial_id) + 1
            print()
            print(f"Search: Running Trial #{self.trial_number}")
            print()

            self.trial_start = datetime.now()
            if self.search_start is None:
                self.search_start = self.trial_start

            self.show_hyperparameter_table(trial)
            print()

    def on_trial_end(self, trial):
        if self.verbose < 1:
            return
        utils.try_clear()

        time_taken_str = self.format_duration(datetime.now() - self.trial_start)
        print(f"Trial {self.trial_number} Complete [{time_taken_str}]")

        if trial.score is not None:
            print(f"{self.oracle.objective.name}: {trial.score}")

        print()
        best_trials = self.oracle.get_best_trials()
        best_score = best_trials[0].score if len(best_trials) > 0 else None
        print(f"Best {self.oracle.objective.name} So Far: {best_score}")

        time_elapsed_str = self.format_duration(
            datetime.now() - self.search_start
        )
        print(f"Total elapsed time: {time_elapsed_str}")

    def show_hyperparameter_table(self, trial):
        template = "{{0:{0}}}|{{1:{0}}}|{{2}}".format(self.col_width)
        best_trials = self.oracle.get_best_trials()
        best_trial = best_trials[0] if len(best_trials) > 0 else None
        if trial.hyperparameters.values:
            print(
                template.format("Value", "Best Value So Far", "Hyperparameter")
            )
            for hp, value in trial.hyperparameters.values.items():
                best_value = (
                    best_trial.hyperparameters.values.get(hp)
                    if best_trial
                    else "?"
                )
                print(
                    template.format(
                        self.format_value(value),
                        self.format_value(best_value),
                        hp,
                    )
                )
        else:
            print("default configuration")

    def format_value(self, val):
        if isinstance(val, (int, float)) and not isinstance(val, bool):
            return f"{val:.5g}"
        val_str = str(val)
        if len(val_str) > self.col_width:
            val_str = f"{val_str[:self.col_width - 3]}..."
        return val_str

    def format_duration(self, d):
        s = round(d.total_seconds())
        d = s // 86400
        s %= 86400
        h = s // 3600
        s %= 3600
        m = s // 60
        s %= 60

        if d > 0:
            return f"{d:d}d {h:02d}h {m:02d}m {s:02d}s"
        return f"{h:02d}h {m:02d}m {s:02d}s"


class SaveBestEpoch(keras.callbacks.Callback):
    """A Keras callback to save the model weights at the best epoch.

    Args:
        objective: An `Objective` instance.
        filepath: String. The file path to save the model weights.
    """

    def __init__(self, objective, filepath):
        super().__init__()
        self.objective = objective
        self.filepath = filepath
        if self.objective.direction == "max":
            self.best_value = float("-inf")
        else:
            self.best_value = float("inf")

    def on_epoch_end(self, epoch, logs=None):
        if not self.objective.has_value(logs):
            # Save on every epoch if metric value is not in the logs. Either no
            # objective is specified, or objective is computed and returned
            # after `fit()`.
            self._save_model()
            return
        current_value = self.objective.get_value(logs)
        if self.objective.better_than(current_value, self.best_value):
            self.best_value = current_value
            self._save_model()

    def _save_model(self):
        # Create temporary saved model files on non-chief workers.
        write_filepath = file_utils.write_filepath(
            self.filepath, self.model.distribute_strategy
        )
        self.model.save_weights(write_filepath)
        # Remove temporary saved model files on non-chief workers.
        file_utils.remove_temp_dir_with_filepath(
            write_filepath, self.model.distribute_strategy
        )


def average_metrics_dicts(metrics_dicts):
    """Averages the metrics dictionaries to one metrics dictionary."""
    metrics = collections.defaultdict(list)
    for metrics_dict in metrics_dicts:
        for metric_name, metric_value in metrics_dict.items():
            metrics[metric_name].append(metric_value)
    averaged_metrics = {
        metric_name: np.mean(metric_values)
        for metric_name, metric_values in metrics.items()
    }

    return averaged_metrics


def _get_best_value_and_best_epoch_from_history(history, objective):
    # A dictionary to record the metric values through epochs.
    # Usage: epoch_metric[epoch_number][metric_name] == metric_value
    epoch_metrics = collections.defaultdict(dict)
    for metric_name, epoch_values in history.history.items():
        for epoch, value in enumerate(epoch_values):
            epoch_metrics[epoch][metric_name] = value
    best_epoch = 0
    for epoch, metrics in epoch_metrics.items():
        objective_value = objective.get_value(metrics)
        # Support multi-objective.
        if objective.name not in metrics:
            metrics[objective.name] = objective_value
        best_value = epoch_metrics[best_epoch][objective.name]
        if objective.better_than(objective_value, best_value):
            best_epoch = epoch
    return epoch_metrics[best_epoch], best_epoch


def convert_to_metrics_dict(results, objective):
    """Convert any supported results type to a metrics dictionary."""
    # List of multiple exectuion results to be averaged.
    # Check this case first to deal each case individually to check for errors.
    if isinstance(results, list):
        return average_metrics_dicts(
            [convert_to_metrics_dict(elem, objective) for elem in results]
        )

    # Single value.
    if isinstance(results, (int, float, np.floating)):
        return {objective.name: float(results)}

    # A dictionary.
    if isinstance(results, dict):
        return results

    # A History.
    if isinstance(results, keras.callbacks.History):
        best_value, _ = _get_best_value_and_best_epoch_from_history(
            results, objective
        )
        return best_value


def validate_trial_results(results, objective, func_name):
    if isinstance(results, list):
        for elem in results:
            validate_trial_results(elem, objective, func_name)
        return

    # Single value.
    if isinstance(results, (int, float, np.floating)):
        return

    # None
    if results is None:
        raise errors.FatalTypeError(
            f"The return value of {func_name} is None. "
            "Did you forget to return the metrics? "
        )

    # objective left unspecified,
    # and objective value is not a single float.
    if isinstance(objective, obj_module.DefaultObjective) and not (
        isinstance(results, dict) and objective.name in results
    ):
        raise errors.FatalTypeError(
            f"Expected the return value of {func_name} to be "
            "a single float when `objective` is left unspecified. "
            f"Recevied return value: {results} of type {type(results)}."
        )

    # A dictionary.
    if isinstance(results, dict):
        if objective.name not in results:
            raise errors.FatalValueError(
                f"Expected the returned dictionary from {func_name} to have "
                f"the specified objective, {objective.name}, "
                "as one of the keys. "
                f"Received: {results}."
            )
        return

    # A History.
    if isinstance(results, keras.callbacks.History):
        return

    # Other unsupported types.
    raise errors.FatalTypeError(
        f"Expected the return value of {func_name} to be "
        "one of float, dict, keras.callbacks.History, "
        "or a list of one of these types. "
        f"Recevied return value: {results} of type {type(results)}."
    )


def get_best_step(results, objective):
    # Average the best epochs if multiple executions.
    if isinstance(results, list):
        return int(
            statistics.mean(
                [get_best_step(elem, objective) for elem in results]
            )
        )

    # A History.
    if isinstance(results, keras.callbacks.History):
        _, best_epoch = _get_best_value_and_best_epoch_from_history(
            results, objective
        )
        return best_epoch

    return 0


def convert_hyperparams_to_hparams(hyperparams):
    """Converts KerasTuner HyperParameters to TensorBoard HParams."""
    hparams = {}
    for hp in hyperparams.space:
        hparams_value = {}
        try:
            hparams_value = hyperparams.get(hp.name)
        except ValueError:  # pragma: no cover
            continue  # pragma: no cover

        hparams_domain = {}
        if isinstance(hp, hp_module.Choice):
            hparams_domain = hparams_api.Discrete(hp.values)
        elif isinstance(hp, hp_module.Int):
            if hp.step is not None and hp.step != 1:
                # Note: `hp.max_value` is inclusive, unlike the end index
                # of Python `range()`, which is exclusive
                values = list(range(hp.min_value, hp.max_value + 1, hp.step))
                hparams_domain = hparams_api.Discrete(values)
            else:
                hparams_domain = hparams_api.IntInterval(
                    hp.min_value, hp.max_value
                )
        elif isinstance(hp, hp_module.Float):
            if hp.step is not None:
                # Note: `hp.max_value` is inclusive, unlike the end index
                # of Numpy's arange(), which is exclusive
                values = np.arange(
                    hp.min_value, hp.max_value + 1e-7, step=hp.step
                ).tolist()
                hparams_domain = hparams_api.Discrete(values)
            else:
                hparams_domain = hparams_api.RealInterval(
                    hp.min_value, hp.max_value
                )
        elif isinstance(hp, hp_module.Boolean):
            hparams_domain = hparams_api.Discrete([True, False])
        elif isinstance(hp, hp_module.Fixed):
            hparams_domain = hparams_api.Discrete([hp.value])
        else:
            raise ValueError(  # pragma: no cover
                f"`HyperParameter` type not recognized: {hp}"
            )

        hparams_key = hparams_api.HParam(hp.name, hparams_domain)
        hparams[hparams_key] = hparams_value

    return hparams
